"""
    DualOptimzer.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import copy
import logging
import numpy as np
from scipy.optimize import minimize
from molass_legacy.Models.ElutionCurveModels import EGHA, EMGA, egha, emga
from molass_legacy._MOLASS.SerialSettings import get_setting
from DualEvaluator import DualEvaluator
import molass_legacy.KekLib.DebugPlot as plt

ADOPTABLE_LIMIT_RATIO       = 0.02
ADOPTABLE_LIMIT_RATIO_YET   = 0.03  # for minor peaks
ADOPTABLE_ADJACENT_RATIO    = 0.2   # not used
OPTIMIZABLE_LIMT_RATIO      = 0.1

class DualPairSelector:
    def __init__(self, curve, fit_recs):
        self.curve = curve
        self.peak_info = curve.peak_info
        self.boundaries = curve.boundaries
        self.fit_recs = fit_recs
        self.fit_top_ys = []
        for rec in fit_recs:
            peak = rec[3]
            y = curve.spline(peak.top_x)
            self.fit_top_ys.append(y)
        self.max_y = curve.max_y
        self.optimize_limit_y = curve.max_y * OPTIMIZABLE_LIMT_RATIO

    def select_one_pair(self):
        pair = None
        if len(self.peak_info) == 1:
            if len(self.fit_recs) == 2:
                # as in Ald
                pair = [0, 1]
        elif len(self.peak_info) == 2:
            if len(self.fit_recs) == 2:
                # as in TSsome2
                pair = [0, 1]
            elif len(self.fit_recs) == 3:
                boundary = self.boundaries[0]
                added_rec = None
                pair = []
                for k, rec in enumerate(self.fit_recs):
                    if rec[0] >= 0:
                        pair.append(k)
                    else:
                        added_rec = rec
                assert added_rec is not None
                func = added_rec[1]
                peak = added_rec[3]
                print( 'peak.top_x=', peak.top_x )
                y = func(peak.top_x)
                if self.curve.is_ignorable_peak(y):
                    pass
                else:
                    assert False
                dist = abs( peak.top_x - boundary )
                print( 'dist=', dist )
                if dist < 10:
                    pass
                else:
                    # as in OA_Ald
                    pair = None
        else:
            pass

        if pair is not None:
            assert len(pair) == 2

        return pair

    def get_nearest_major_peak_index(self, i):
        this_top_x = self.fit_recs[i][3].top_x
        min_dist = None
        nearest = None
        for k, rec in enumerate(self.fit_recs):
            if k == i or rec[0] < 0:
                continue
            dist = abs(self.fit_recs[i][3].top_x - this_top_x)
            if min_dist is None or dist < min_dist:
                min_dist = dist
                nearest = k
        return nearest

    def is_adoptable(self, i, debug=False):

        peak_ratio = self.fit_top_ys[i]/self.max_y
        if debug:
            print([i], 'peak_ratio=', peak_ratio)

        if peak_ratio < ADOPTABLE_LIMIT_RATIO:
            if debug:
                print([i], 'is not adoptable: (1)')
            return False

        if len(self.fit_recs) == 1:
            if debug:
                print([i], 'is adoptable: (2)')
            return True

        rec = self.fit_recs[i]
        if rec[0] >= 0:
            if debug:
                print([i], 'is adoptable: (3)')
            return True

        if peak_ratio > ADOPTABLE_LIMIT_RATIO_YET:
            if debug:
                print([i], 'is adoptable: (4)')
            return True

        x = rec[3].top_x

        # get the adjacent major peak func
        k = self.get_nearest_major_peak_index(i)
        if k is None:
            if debug:
                print([i], 'is not adoptable: (5)')
            return False

        if False:
            # TODO: remove this (seems unnecessary)

            func_ = self.fit_recs[k][1]
            foot_ratio = func_(x)/self.max_y

            if debug:
                print([i,k], 'x=', x, 'foot_ratio=', foot_ratio)
            # print( [i, k], 'adjacent y=', y,  )
            # adjacent y= 9.87e-06 8.00e-06 for th 1st resid peak in Ald

            if foot_ratio > ADOPTABLE_LIMIT_RATIO**2:
                # as in the 2nd resid peak in Mar07
                if debug:
                    print([i], 'is adoptable: (6) foot_ratio=', foot_ratio)
                return True

        if debug:
            print([i], 'is not adoptable: (7)')

        return False

    def is_optimizable_rec(self, rec):
        if rec[0] >= 0:
            return True

        peak = rec[3]
        # print( 'peak.top_y=', peak.top_y, 'peak.sign=', peak.sign )
        if peak.sign < 0:
            return False

        if peak.top_y >= self.optimize_limit_y:
            return True

        return False

    def is_optimizable_pair(self, i, j, debug=False):
        assert i+1 == j or i+2 == j

        rec1 = self.fit_recs[i]
        rec2 = self.fit_recs[j]
        if not self.is_optimizable_rec(rec1) or not self.is_optimizable_rec(rec2):
            return False

        peak1 = rec1[3]
        peak2 = rec2[3]

        if j == i+1:
            # print('optimize_limit_y=', self.optimize_limit_y)
            for peak in [peak1, peak2]:
                x = peak.top_x
                y = self.curve.spline(x)
                if y < self.optimize_limit_y:
                    return False
        else:
            # as in 20160227
            pass

        if j == i+2:
            rec_ = self.fit_recs[i+1]
            func1 = rec1[1]
            func2 = rec2[1]
            func_ = rec_[1]
            peak_ = rec_[3]
            px = peak_.top_x
            py1 = func1(px)
            py_ = func_(px)
            py2 = func2(px)
            optimizable = py1/py_ > 0.5 and py2/py_ > 0.5

            if debug:
                print( 'optimizable', optimizable )
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("is_optimizable_pair debug: i=%d, j=%d" % (i, j))
                    ax.plot(self.curve.x, self.curve.y)
                    for peak in [peak1, peak2]:
                        x = peak.top_x
                        y = self.curve.spline(x)
                        ax.plot(x, y, 'o', color='red')

                    for y in [py1, py_, py2]:
                        ax.plot(px, y, 'o', color='yellow')
                    plt.show()

            return optimizable

        return True

    def compute_y_for_opt(self, x, y, i, j):
        y_ = copy.deepcopy(y)
        for k, rec in enumerate(self.fit_recs):
            if i <= k and k <= j:
                continue

            func = rec[1]
            y_ -= func(x)
        return y_

class DualOptimzer:
    def __init__(self, x, y, params1, params2, tau_hint_pair=None):
        assert type(params1) is np.ndarray and type(params2) is np.ndarray
        self.logger = logging.getLogger(__name__)
        self.x = x
        self.y = y
        self.init_params = np.concatenate([params1, params2])
        """
            editor_frame.recompute_decomposition()
                hints_dict = self.get_param_hints_dict()
                decompose_elution_better(..., hints_dict, ...)
                    decompose_elution_xray_to_uv(..., hints_dict, ...)
                        decompose_elution_impl(..., hints_dict, ...)
                            get_dual_optimized_info( ..., hints_dict  , ... )
                    decompose_elution_uv_to_xray(..., hints_dict, ...)
                        decompose_elution_impl(..., hints_dict, ...)
                            get_dual_optimized_info( ..., hints_dict  , ... )

            get_dual_optimized_info( ..., hints_dict  , ... )
                hints_dict → tau_hint_pair
                dopt = optimizer_class(..., tau_hint_pair=tau_hint_pair)
        """
        # self.tau_hint_pair = tau_hint_pair
        self.tau_bound_ratio = get_setting("TAU_BOUND_RATIO")
        # self.debug_plot("init", self.init_params)

    def debug_plot(self, title, params):
        # self.func must be given by a Subclass
        print( title, params )
        x = self.x
        y = self.y
        ax = plt.gca()
        ax.cla()
        ax.set_title(title)
        ax.plot(x,y,color='orange')
        arg1 = self.init_params[0:5]
        ax.plot(x, self.func(x,*arg1))
        arg2 = self.init_params[5:10]
        ax.plot(x, self.func(x,*arg2))
        plt.show()

    def obj_func(self, p):
        # print('obj_func: p', p)
        return np.sum(( self.func(self.x, *p[0:5]) + self.func(self.x, *p[5:10]) - self.y )**2)

    def optimize(self):
        # self.func must be given by a Subclass

        tau_bound_ratio = self.tau_bound_ratio

        #  0  1   2      3    4    5  6   7     8    9
        #  h  mu  sigma  tau  a    h  mu  sigma tau  a
        cons = [
                {'type': 'ineq', 'fun': lambda x:  x[2]*tau_bound_ratio - abs(x[3]) },  # sigma*tau_bound_ratio - abs(tau) >= 0
                {'type': 'ineq', 'fun': lambda x:  x[7]*tau_bound_ratio - abs(x[8]) },  # sigma*tau_bound_ratio - abs(tau) >= 0
            ]

        self.logger.info("minimizing with tau_bound_ratio=%g", tau_bound_ratio)
        res = minimize(self.obj_func, self.init_params, method='Nelder-Mead', constraints=cons)
        # print('opt p=', res.x)
        # self.debug_plot("optimized", res.x)
        return res

    def get_evaluators(self, res):
        # self.eval_class must be given by a Subclass
        return [ DualEvaluator(self.model, p) for p in [res.x[0:5], res.x[5:10]] ]

    def get_optrec(self, fitrec, res, k):
        rec = copy.deepcopy(fitrec)
        rec.set_evaluator(DualEvaluator(self.model, res.x[5*k:5*(k+1)]))
        return rec

    def get_avgrec(self, fitrec, res1, res2):
        rec = copy.deepcopy(fitrec)
        avg_params = (res1.x[5:10] + res2.x[0:5])/2
        rec.set_evaluator(DualEvaluator(self.model, avg_params))
        return rec

class EmgaOptimzer(DualOptimzer):
    def __init__(self, x, y, params1, params2, tau_hint_pair=None):
        self.model = EMGA()
        self.model_name = "EMGA"
        self.func = emga
        DualOptimzer.__init__(self, x, y, params1, params2, tau_hint_pair)

class EghaOptimzer(DualOptimzer):
    def __init__(self, x, y, params1, params2, tau_hint_pair=None):
        self.model = EGHA()
        self.model_name = "EGHA"
        self.func = egha
        DualOptimzer.__init__(self, x, y, params1, params2, tau_hint_pair)

def get_dual_optimized_info_old(optimizer_class, x_curve, x, y, fit_recs):
    dps = DualPairSelector(x_curve, fit_recs)
    dual_pair = dps.select_one_pair()
    print( 'dual_pair=', dual_pair )

    if dual_pair is None:
        fit_recs_ = fit_recs
    else:
        i1, i2 = dual_pair
        dopt = optimizer_class(x, y, fit_recs[i1][1].params, fit_recs[i2][1].params)
        res = dopt.optimize()
        dual_evals = dopt.get_evaluators(res)
        fit_recs_ = []
        for k, rec in enumerate([fit_recs[i1], fit_recs[i2]]):
            rec_ = copy.deepcopy(rec)
            rec_[1] = dual_evals[k]
            fit_recs_.append(rec_)

    return fit_recs_

def get_dual_optimized_info(optimizer_class, x_curve, x, y, fit_recs, hints_dict, logger, debug=False):

    # print('fit_recs=', fit_recs)

    dps = DualPairSelector(x_curve, fit_recs)
    ret_recs = []
    ign_recs = []

    n = len(fit_recs)
    pending_res = None
    i = 0
    for k in range(n):
        if i >= n:
            break

        fitrec1 = fit_recs[i]
        if hints_dict is None:
            hint1 = None
        else:
            xkey1 = fitrec1[3].get_xkey()
            hint1 = hints_dict.get(xkey1)
        if i+2 < n and dps.is_optimizable_pair(i, i+2):
            j = i+2
        elif i+1 < n and dps.is_optimizable_pair(i, i+1):
            j = i+1
        else:
            j = None

        if debug:
            from OptRecsUtils import debug_plot_opt_recs
            print('n=', n, 'i=', i, 'j=', j)
            debug_plot_opt_recs(x_curve, fit_recs, title="get_dual_optimized_info: [%d] i=%d, j=%s" % (k, i, str(j)) )

        if j is None:
            if pending_res is None:
                if dps.is_adoptable(i, debug=debug):
                    ret_recs.append(fitrec1)
                    if debug: print( [i], 'adopted' )
                else:
                    ign_recs.append(fitrec1)
                    if debug: print( [i], 'ignored' )
            else:
                rec = dopt.get_optrec(fitrec1, pending_res, 1)
                ret_recs.append(rec)
                pending_res = None
                if debug: print( [i], 'adopted pending' )
        else:
            fitrec2 = fit_recs[j]
            if hints_dict is None:
                hint2 = None
            else:
                xkey2 = fitrec2[3].get_xkey()
                hint2 = hints_dict.get(xkey2)
            y_ = dps.compute_y_for_opt(x, y, i, j)
            tau_hint_pair = ( hint1, hint2 )
            dopt = optimizer_class(x, y_, fitrec1[1].param_values, fitrec2[1].param_values, tau_hint_pair=tau_hint_pair)
            res = dopt.optimize()
            if pending_res is None:
                rec = dopt.get_optrec(fitrec1, res, 0)
                if debug: print( [i], 'adopted without pending' )
            else:
                rec = dopt.get_avgrec(fitrec1, pending_res, res)
                if debug: print( [i], 'adopted with pending' )

            ret_recs.append(rec)
            if rec[0] < 0:
                # don't let pend minor peaks
                pending_res = None
            else:
                pending_res = res

        if j is None:
            i += 1
        else:
            i = j

    logger.info( 'dual optimizer produced %d records ignoring %d records.' % ( len(ret_recs), len(ign_recs) ) )
    return ret_recs

def get_assert_info(curve, x, y, opt_recs):
    from molass_legacy.KekLib.BasicUtils import Struct
    peaks=[ rec[3].get_assert_info()  for rec in opt_recs ]

    y_ = copy.deepcopy(y)
    for rec in opt_recs:
        func = rec[1]
        y_ -= func(x)

    resid = np.abs(y_)
    navgdiff = np.sum(resid)/len(y)/curve.max_y
    nmaxdiff = np.max(resid)/curve.max_y

    return Struct( peaks=peaks, navgdiff=navgdiff, nmaxdiff=nmaxdiff )



"""
    Tentative.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Models.ElutionModelUtils import get_xies_from_height_ratio

ENABLE_TAU_BOUND = False

class Model:
    def __init__(self, func=None, **kwargs):
        self.func = func
        self.name = "Model(%s)" % self.__class__.__name__
        self.delayed = kwargs.pop("delayed", False)
        self.tau_bound_ratio = get_setting("TAU_BOUND_RATIO")

    def get_name(self):
        return self.name

    def is_delayed(self):
        return self.delayed

    def is_traditional(self):
        return True

    def set_delayed_off(self):
        self.delayed = False

    def make_params(self, **kwargs):
        assert False

    def set_param_hint(self, pname, **kwargs):
        assert False

    def get_param_hints(self, pname):
        # remove this
        return None

    def fit(self, y, params, x=None, method=None):
        assert x is not None

        if method == "least_squares":
            # used in EmgPeak.py
            method = None

        def objective(p):
            try:
                y_ = self.func(x, *p)           # emga may return NaN
                chisqr = np.sum((y_ - y)**2)
            except:
                chisqr = np.inf
            return chisqr


        if ENABLE_TAU_BOUND:
            tau_bound_ratio = self.tau_bound_ratio

            cons = [
                    {'type': 'ineq', 'fun': lambda x:  x[2]*tau_bound_ratio - abs(x[3]) },  # sigma*tau_bound_ratio - abs(tau) >= 0
                    ]

            if method is None:
                method="Nelder-Mead"
        else:
            cons = None

        ret = minimize(objective, params, method=method, constraints=cons)

        chisqr = objective(ret.x)   # chisqr is referenced in ElutionDecomposer
        return Struct(params=ret.x, chisqr=chisqr)

    def eval(self, params, x=None):
        assert x is not None

        return self.func(x, *params)

    def __call__(self, x, params):
        return self.func(x, *params)

    def get_peaktop_xy(self, x, params):
        # this is a fallback. should be overridden if a faster implementation is available
        y = self.func(x, *params)
        m = np.argmax(y)
        return x[m], y[m]

    def get_proportions(self, x, params_array):
        areas = []
        for params in params_array:
            cy = self.func(x, *params)
            areas.append(np.sum(cy))
        return np.array(areas)/np.sum(areas)

    def get_range(self, x, y, alpha=0.1, max_y=None, debug=False):
        # task: unify the alpha value with the same meaning in other methods
        return get_xies_from_height_ratio(alpha, x, y, max_y=max_y, debug=debug)

    def guess_multiple(self, x, y, num_peaks, debug=False):
        # temporary support only for trditional models
        # task: add support for advanced models
        from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks  # placed here to avoid circular import
        assert self.get_name() in ["EGHA", "EMGA"] 
        return np.array(recognize_peaks(x, y, exact_num_peaks=num_peaks, affine=True, model=self, debug=debug))
    
    def guess_a_peak_with_prop(self, x, y, prop, **kwargs):
        raise NotImplementedError()     # should be overridden
    
    def guess_binary_peaks(self, x, y, p1, p2, **kwargs):
        raise NotImplementedError()     # should be overridden
    
    def guess_from_the_other(self, x, y, other_y, params_array):
        """
        This is not expected to work for advanced, i.e. non-traditional, models.
        """
        scale = np.sum(y)/np.sum(other_y)
        new_params_list = []
        for i, params in enumerate(params_array):
            new_params = params.copy()
            new_params[0] *= scale
            new_params_list.append(new_params)
        return np.array(new_params_list)


"""
    SerialData.py

    連続測定データ

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""

import os
import copy
from bisect import bisect_right
import numpy    as np
import re
import threading
import logging
import inspect
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy._MOLASS.SerialSettings import get_setting
from DevSettings import get_dev_setting
from molass_legacy.Elution.CurveUtils import get_xray_elution_vector
from molass_legacy.Trimming import FlangeLimit
from molass_legacy.UV.Absorbance import Absorbance
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from ScatteringBaseUtil import apply_baseline_correction_impl
from InputSmootherAveraging import IntensitySmootherAveraging, ConcentrationSmootherAveraging
from molass_legacy.Test.TesterLogger import put_to_tester_log_queue, write_to_tester_log
from SerialDataUtils import load_intensity_files, load_uv_array
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker

DEBUG = False
DEBUG_PLOT = False

REQUIRED_MIN_QUALITY            = 0.5
NUM_POINTS_DISCOMFORT_DETECT    = 10
REQUIRED_SERIAL_NUM_FILES       = 20

class SerialData:
    def __init__( self, conc_folder, intensity_folder,
                    conc_file=None,
                    data_info=None,
                    orig_info=None,
                    copy_info=None,
                    pre_recog=None,
                    do_mapping=True,
                    debug=False,
                    ):
        self.debug  = debug
        self.do_mapping = do_mapping
        self.q_limit_index  = None
        self.usable_slice   = None          # to be removed
        self.num_curves_averaged = False
        self.intensity_reduced  = False
        self.absorbance = None
        self.ready      = False
        self.initial_std_diff   = None
        self.current_std_diff   = None
        self.has_adjusted_array = False
        self.baseline_corrected = False
        self.corrected_base = None
        self.scattering_baseline_discomfort  = None
        self.mc_vector = None
        self.conc_factor = None
        self.xray_transformed_conc = None
        self.logger         = logging.getLogger( __name__ )
        self.data_info      = data_info
        self.orig_info      = orig_info
        self.copy_info      = copy_info
        self.cd_info = None
        self.set_offsets()
        # self.logger.info('copy_info' + str(copy_info))
        self.logger.info('copy_info from %s', 'None' if copy_info is None else copy_info[0].get_id_info() )
        self.top_x_pair     = self.get_original_peak_top_x_pair()
        self.pre_recog      = pre_recog
        self.helper_info    = None
        self.has_excluded_xray_elutions = False
        self.excluded_set   = set()     # better control info
        self.elutin_slice   = None
        self.cd_slice       = None
        self.mtd_elution    = None

        if data_info is None:
            assert False
            self.read_thread = threading.Thread(
                                    target=self.read,
                                    name='ReadThread',
                                    args=[  conc_folder,
                                            intensity_folder,
                                            conc_file,
                                         ],
                                    )
            self.read_thread.start()
        else:
            self.set_data( data_info, conc_file )

    def __del__(self):
        print("SerialData.__del__ ok")

    def set_offsets(self):
        if self.copy_info is None:
            self.uv_i0 = 0
            self.uv_j0 = 0
            self.xr_i0 = 0
            self.xr_j0 = 0
        else:
            uv_restrict = self.copy_info[1]
            if uv_restrict is None:
                self.uv_i0 = 0
                self.uv_j0 = 0
            else:
                elution, wavelength = uv_restrict
                self.uv_i0 = 0 if wavelength is None else wavelength.start
                self.uv_j0 = 0 if elution is None else elution.start

            xray_restrict = self.copy_info[2]
            if xray_restrict is None:
                self.xr_i0 = 0
                self.xr_j0 = 0
            else:
                elution, angle = xray_restrict
                self.xr_i0 = 0 if angle is None else angle.start
                self.xr_j0 = 0 if elution is None else elution.start

    def set_offsets_from_md(self, md):
        uv = md.uv
        self.uv_i0 = uv.i_slice.start
        self.uv_j0 = uv.j_slice.start

        xr = md.xr
        self.xr_i0 = xr.i_slice.start
        self.xr_j0 = xr.j_slice.start

    def set_prerecog_proxy(self, md):
        from PrerecogProxy import PrerecogProxy
        self.pre_recog = PrerecogProxy(md)

    def read( self, conc_folder, intensity_folder, conc_file ):
        # print( 'reading: ', intensity_folder,  conc_file)
        self.logger.info( 'loading by SerialData' )
        self.load_conc_array( conc_folder, conc_file=conc_file )
        self.load_intensity_array(intensity_folder)
        # self.make_adjusted_array() should be call on button-press either by "Figure" or "Run"

    def wait_until_ready( self ):
        if self.data_info is None:
            assert False
            self.read_thread.join()
            if self.absorbance is None:
                error_msg = 'absorbance is not ready'
                write_to_tester_log( error_msg )
                raise RuntimeError( error_msg )
        else:
            pass
        self.ready  = True

    def load_conc_array( self, conc_folder, conc_file=None ):
        data_array, lvector, conc_file = load_uv_array( conc_folder, conc_file )
        self.conc_array = data_array
        self.lvector    = lvector
        self.conc_file  = conc_file

    def load_intensity_array(self, in_folder):
        if not os.path.exists( in_folder ):
            raise Exception( in_folder + ' does not exist!'  )

        data_array, datafiles = load_intensity_files( in_folder, self.logger )

        self.num_datafiles = len( datafiles )
        self.datafiles = datafiles
        self.set_intensity_array( data_array )

    def set_data( self, data_info, conc_file ):
        datafiles, xray_array, uv_array, lvector, col_header, mtd_elution = data_info

        # uv-data
        self.conc_array = uv_array
        self.lvector    = lvector
        self.conc_file  = conc_file
        self.col_header = col_header
        self.mtd_elution = mtd_elution

        # xray-data
        self.num_datafiles = len( datafiles )
        self.datafiles  = datafiles
        data_num_re = re.compile( r'_(\d+)' )
        _, first_file = os.path.split( self.datafiles[0] )
        m = data_num_re.search( first_file )
        if m:
            self.start_file_no  = int( m.group(1) )
        else:
            self.start_file_no  = 0
        self.set_intensity_array( xray_array )

        orig_top_x = None if self.top_x_pair else self.top_x_pair[0]
        self.absorbance = Absorbance( lvector, uv_array, self.xray_curve, col_header=col_header, orig_top_x=orig_top_x )

        self.ready  = True

    def get_data_shapes( self ):
        return self.intensity_array.shape, self.conc_array.shape

    def _get_analysis_copy_impl(self, given_pre_recog, return_also_new_prerecog=False, debug=False):
        datafiles   = self.datafiles
        xray_array  = self.intensity_array
        uv_array    = self.conc_array
        lvector     = self.lvector
        col_header  = self.col_header

        xr_restrict_list = get_setting('xr_restrict_list')
        if xr_restrict_list is None:
            # usable_slice = self.get_usable_slice()
            # xray_array_ = xray_array[:,usable_slice,:]
            xray_array_ = xray_array
            datafiles_ = datafiles
        else:
            # print('xr_restrict_list=', xr_restrict_list)
            elution_restrict = xr_restrict_list[0]
            if elution_restrict is None:
                eslice = slice(None, None)
            else:
                eslice = slice(elution_restrict.start, elution_restrict.stop)
            angle_restrict = xr_restrict_list[1]
            if angle_restrict is None:
                sslice = slice(None, None)
            else:
                sslice = slice(angle_restrict.start, angle_restrict.stop)
            xray_array_ = xray_array[eslice,sslice,:]
            datafiles_ = datafiles[eslice]
            self.logger.info('xray_array has been restricted from %s to %s' % (str(xray_array.shape), str(xray_array_.shape)))

        uv_restrict_list = get_setting('uv_restrict_list')
        col_header_ = col_header
        if uv_restrict_list is None:
            uv_array_   = uv_array
            lvector_    = lvector
        else:
            elution_restrict = uv_restrict_list[0]
            if elution_restrict is None:
                eslice = slice(None, None)
            else:
                eslice = slice(elution_restrict.start, elution_restrict.stop)
            wl_restrict = uv_restrict_list[1]
            if wl_restrict is None:
                wslice = slice(None, None)
            else:
                wslice = slice(wl_restrict.start, wl_restrict.stop+1)
            lvector_ = lvector[wslice]
            uv_array_ = uv_array[wslice, eslice]
            if col_header is not None:
                col_header_ = col_header[eslice]
            self.logger.info('uv_array has been restricted from %s to %s' % (str(uv_array.shape), str(uv_array_.shape)))

        # data_info = copy.deepcopy( [ datafiles_, xray_array_, uv_array_, lvector, col_header_, self.mtd_elution ] )
        data_info = copy.deepcopy( [ datafiles_, xray_array_, uv_array_, lvector_, col_header_] )
        data_info += [self.mtd_elution]
        analysis_copy = SerialData( None, None,
                            conc_file=self.conc_file,
                            data_info=data_info,
                            pre_recog=given_pre_recog,
                            copy_info=[self, uv_restrict_list, xr_restrict_list],
                            )
        self.logger.info("serialdata analysis copy has been made with trimming info UV %s and Xray %s", str(uv_restrict_list), str(xr_restrict_list))
        self.logger.info( 'serialdata analysis copy %s made with xray_array.shape=%s from %s', analysis_copy.get_id_info(), str(analysis_copy.intensity_array.shape), self.get_id_info() )

        if return_also_new_prerecog:
            from importlib import reload
            import molass.DataUtils.ForwardCompat
            reload(molass.DataUtils.ForwardCompat)
            from molass.DataUtils.ForwardCompat import convert_to_trimmed_prerecog
            new_pre_recog = convert_to_trimmed_prerecog(given_pre_recog, uv_restrict_list, xr_restrict_list, debug=debug)
            return analysis_copy, new_pre_recog
        else:
            return analysis_copy

    def get_original_peak_top_x_pair(self):
        if self.copy_info is None:
            return [None, None]

        orig_sd, uv_restrict_list, xr_restrict_list = self.copy_info

        ret_top_x_list = []
        a_curve = orig_sd.absorbance.a_curve
        x_curve = orig_sd.xray_curve

        for curve, list_ in zip([a_curve, x_curve], [uv_restrict_list, xr_restrict_list]):

            peak_top_x = curve.peak_top_x

            if list_ is None:
                ret_top_x = peak_top_x
            else:
                elution_restrict = list_[0]
                if elution_restrict is None:
                    delta_x = 0
                else:
                    delta_x = elution_restrict.start
                ret_top_x = peak_top_x - delta_x

            ret_top_x_list.append(ret_top_x)

        return ret_top_x_list

    def get_copy( self, pre_recog=None ):
        datafiles   = self.datafiles
        xray_array  = self.intensity_array
        uv_array    = self.conc_array
        lvector     = self.lvector
        col_header  = self.col_header
        # data_info = copy.deepcopy( [ datafiles, xray_array, uv_array, lvector, col_header, self.mtd_elution ] )
        data_info = copy.deepcopy( [ datafiles, xray_array, uv_array, lvector, col_header])
        data_info += [self.mtd_elution]
        sd_ = SerialData( None, None, conc_file=self.conc_file, data_info=data_info, pre_recog=pre_recog, copy_info=self.copy_info )
        sd_.baseline_corrected = self.baseline_corrected
        sd_.corrected_base = self.corrected_base
        sd_.cd_info = self.cd_info
        sd_.excluded_set = self.excluded_set
        return sd_

    def get_exec_copy( self, mapper, conc_factor=None ):
        if conc_factor is None:
            conc_factor = self.conc_factor
        # note that conc_factor must be specified for the first copy
        assert conc_factor is not None

        exec_copy = self.get_copy(pre_recog=self.pre_recog)
        exec_copy.set_mc_vector( mapper, conc_factor )

        self.logger.info( 'serialdata exec copy %s made with xray_array.shape=%s from %s', exec_copy.get_id_info(), str(exec_copy.intensity_array.shape), self.get_id_info() )
        return exec_copy

    def get_exec_copy_inside_of_flowchanges( self, mapper, conc_factor ):
        start, stop = mapper.flow_changes
        if start is None:
            start = 0
        if stop is None:
            stop  = start + self.num_datafiles

        elutin_slice = slice(start, stop)
        usable_slice = self.get_usable_slice()

        datafiles   = self.datafiles[elutin_slice]
        xray_array  = self.intensity_array[elutin_slice,usable_slice]
        uv_array    = self.conc_array
        lvector     = self.lvector
        col_header  = self.col_header
        mtd_elution = self.mtd_elution

        data_info   = copy.deepcopy( [ datafiles, xray_array, uv_array, lvector, col_header, mtd_elution ] )
        exec_copy = SerialData( None, None, conc_file=self.conc_file, data_info=data_info, copy_info=self.copy_info )
        exec_copy.set_mc_vector( mapper, conc_factor, elutin_slice=elutin_slice )
        exec_copy.elutin_slice = elutin_slice

        return exec_copy

    def get_v2_copy(self, pre_recog):
        start, stop = pre_recog.get_angle_range()
        datafiles   = self.datafiles
        xray_array  = self.intensity_array[:,start:stop,:]
        uv_array    = self.conc_array
        lvector     = self.lvector
        col_header  = self.col_header
        data_info = copy.deepcopy( [ datafiles, xray_array, uv_array, lvector, col_header])
        data_info += [self.mtd_elution]
        sd_ = SerialData( None, None, conc_file=self.conc_file, data_info=data_info, pre_recog=pre_recog, copy_info=self.copy_info )
        sd_.baseline_corrected = self.baseline_corrected
        sd_.corrected_base = self.corrected_base
        sd_.cd_info = self.cd_info
        sd_.excluded_set = self.excluded_set
        return sd_

    def get_xray_elution_vector( self ):
        ivector_, slice_ = get_xray_elution_vector( self.qvector, self.intensity_array )
        self.xray_slice = slice_
        self.xray_index = (slice_.start + slice_.stop)//2
        self.xr_index = self.xray_index     # for forward compatibility 
        return ivector_

    def is_serial( self ):
        return len(self.ivector) >= REQUIRED_SERIAL_NUM_FILES

    def set_ivector_etcetera(self, debug=False):
        self.ivector = self.get_xray_elution_vector()
        if not self.is_serial():
            return

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("set_ivector_etcetera")
                ax.plot(self.ivector, label="ivector")
                ax.legend()
                fig.tight_layout()
                plt.show()

        orig_top_x = None if self.top_x_pair is None else self.top_x_pair[1]
        self.xray_curve = ElutionCurve( self.ivector, orig_top_x=orig_top_x )

    def update_xray_elution_info(self, ivector, xray_slice):
        self.ivector = ivector
        self.xray_slice = xray_slice
        self.xray_index = (xray_slice.start + xray_slice.stop)//2
        self.xr_index = self.xray_index     # for forward compatibility 
        orig_top_x = None if self.top_x_pair is None else self.top_x_pair[1]
        self.xray_curve = ElutionCurve( self.ivector, orig_top_x=orig_top_x )

    def get_uv_elution_vector( self ):
        std_wvlen   = get_setting( 'absorbance_picking' )
        a_vector, i = self.absorbance.get_vector_at( std_wvlen )
        return a_vector

    def get_uv_curve( self ):
        return self.absorbance.a_curve

    def get_xray_curve( self ):
        return self.xray_curve

    def get_xr_curve( self ):
        return self.xray_curve

    def set_intensity_array( self, new_intensity_array ):
        self.intensity_array = new_intensity_array
        self.qvector = self.intensity_array[0, :, 0]
        self.jvector = np.arange( self.intensity_array.shape[0] )
        self.set_ivector_etcetera()

    def apply_data_reduction( self ):
        if self.intensity_reduced:
            # do not repeat reduction
            return

        if get_dev_setting( 'intensity_reduction' ) == 0:
            return

        reduction_cycle = get_dev_setting( 'reduction_cycle' )
        reduction_start = get_dev_setting( 'reduction_start' )
        reduced_array = self.intensity_array[ :, slice( reduction_start, None, reduction_cycle ), : ]
        print( 'Intesity data reduced by ', reduction_cycle, reduction_start )
        self.set_intensity_array( reduced_array )

        self.intensity_reduced = True

    def get_usable_slice( self ):
        self.get_usable_q_limit()
        return self.usable_slice

    def get_usable_q_limit( self, debug=False):
        self.logger.warning('get_usable_q_limit is deprecated. use FlangeLimit instead.')

        if self.usable_slice is not None:
            return self.q_limit_index

        if self.is_serial():
            if self.pre_recog is not None:
                return self.pre_recog.flange_limit

            matrix_data = self.intensity_array[:,:,1].T
            error_data = self.intensity_array[:,:,2].T
            fl = FlangeLimit(matrix_data, error_data, self.xray_curve, self.qvector)
            limit = fl.get_limit(debug=debug)
        else:
            limit = None

        q_limit = None if limit is None else self.qvector[limit]
        write_to_tester_log( 'q_limit_index=%s, q_limit=%s\n' % (str(limit), str(q_limit)) )

        self.q_limit_index = limit
        self.usable_slice = slice( 0, limit )
        return limit

    def get_around_slice( self, vector, val, num ):
        num_half    = num//2
        num_r       = num%2
        i = bisect_right( vector, val )
        return max( 0, i - num_half ), min( len(self.qvector), i + num_half + num_r )

    def update_mapping_info( self, mapper ):
        if self.initial_std_diff is None:
            self.initial_std_diff = mapper.std_diff
        self.current_std_diff = mapper.std_diff

    def set_mc_vector( self, mapper, conc_factor, elutin_slice=None ):
        # TODO: to be removed
        conc_vector = mapper.get_conc_vector( conc_factor )
        if elutin_slice is None:
            mc_vector = conc_vector
        else:
            mc_vector = conc_vector[elutin_slice]
        self.mc_vector  = mc_vector
        self.conc_factor = conc_factor      # TODO: better place to hold this info

    def exclude_intensities( self, to_be_excluded ):
        from molass.DataUtils.AnomalyHandlers import remove_bubbles_impl
        self.wait_until_ready()
        print( 'to_be_excluded=', to_be_excluded )

        remove_bubbles_impl(self.intensity_array, to_be_excluded, self.excluded_set)

        self.has_excluded_xray_elutions = True
        self.set_ivector_etcetera()
        self.logger.info( 'Xray data at elution points ' + str(to_be_excluded) + ' have been removed and interpolated.' )

    def apply_baseline_correction( self, mapped_info, basic_lpm=False, progress_cb=None, return_base=False, debug_obj=None  ):
        frame = inspect.stack()[1]
        self.logger.info("apply baseline_correction called from %s(%d) with basic_lpm=%s", frame.filename, frame.lineno, str(basic_lpm))
        self.logger.info("id:%s, code_context:%s", str(id(self)),''.join(frame.code_context)[:-1])

        debug = False
        if debug:
            xr_array = self.intensity_array.copy()

        base = apply_baseline_correction_impl(
            self.jvector, self.qvector, self.intensity_array,
            mapped_info, basic_lpm=basic_lpm,
            index=self.xray_index,
            ecurve=self.xray_curve,
            progress_cb=progress_cb,
            return_base=return_base, logger=self.logger, debug_obj=debug_obj )

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.DataStructure.MatrixData import simple_plot_3d
            qv = self.qvector

            with plt.Dp():
                ey1 = get_xray_elution_vector(qv, xr_array)[0]
                ey2 = get_xray_elution_vector(qv, self.intensity_array)[0]
                fig, ax = plt.subplots()
                ax.set_title("apply_baseline_correction: elution curves")
                ax.plot(ey1, label="original")
                ax.plot(ey2, label="corrected")
                ax.legend()
                fig.tight_layout()
                plt.show()

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5), subplot_kw=dict(projection="3d"))
                fig.suptitle("apply_baseline_correction: 3d views")
                simple_plot_3d(ax1, xr_array[:,:,1].T, x=qv)
                simple_plot_3d(ax2, self.intensity_array[:,:,1].T, x=qv)
                fig.tight_layout()
                plt.show()

        self.baseline_corrected = True
        self.corrected_base = base
        try:
            self.set_ivector_etcetera(debug=debug)
        except:
            # not fatal as in the degraded case of 20161124
            etb = ExceptionTracebacker()
            self.logger.warning( etb )

        return base

    def compute_scattering_baseline_discomfort( self ):
        if self.scattering_baseline_discomfort is not None:
            return self.scattering_baseline_discomfort

        size    = len(self.intensity_array)
        j       = self.jvector

        param_list = []
        i_end = (self.xray_slice.start + self.xray_slice.stop) // 2 # i.e., average(start, stop) * 2
        for i in np.linspace( 0, i_end, NUM_POINTS_DISCOMFORT_DETECT, dtype=int ):
            y = self.intensity_array[:,i,1]
            # giving curve=self.xray_curve to improve failed LPM cases such as 20181127
            sbl = ScatteringBaseline( y, curve=self.xray_curve )
            A, B = sbl.solve()
            # print( [ i, self.qvector[i], A, B ] )
            M = np.percentile( y, 95 )
            baseline = A*j + B
            integ   = np.sum( baseline )
            integ_a = np.sum( np.abs( baseline ) )
            param_list.append( [A, B, M, size, integ, integ_a] )

        param_array = np.array( param_list )
        average = np.average( param_array[:, 0:4], axis=0 )
        # print( 'average base params=', average )
        sum_    = np.sum( param_array[:, 4:], axis=0 ) / NUM_POINTS_DISCOMFORT_DETECT
        # print( 'sum_=', sum_ )
        # print( 'integrals=', param_array[:, 4] )
        record  = np.hstack( [ average, sum_ ] )
        recstr = '[ ' + ' '.join( [ str(v) for v in record ] ) + ' ]'
        """
            average( A, B, M, size ), sum( integ, integ_a )
        """
        put_to_tester_log_queue( 'baseline discomfort record=%s\n' % recstr  )

        integ_a_sum   = sum_[1]
        self.scattering_baseline_discomfort = integ_a_sum / average[2] / size

        return self.scattering_baseline_discomfort

    def get_averaged_data( self, num_curves_averaged ):

        # Intensity Smoothing
        smoother = IntensitySmootherAveraging( self.intensity_array, num_curves_averaged )
        indeces = np.arange( 0, self.intensity_array.shape[0] )     # these indeces start at 0, not at self.xr_j0
        averaged_intensity_array, average_slice_array = smoother( indeces, return_numpy_ndarray=True )
        self.num_curves_averaged        = num_curves_averaged
        self.averaged_intensity_array   = averaged_intensity_array
        self.average_slice_array        = average_slice_array

        # Concetration Smoothing
        if self.mc_vector is None:
            self.averaged_c_vector = None
        else:
            cvector_smoother = ConcentrationSmootherAveraging( self.mc_vector, num_curves_averaged )
            self.averaged_c_vector =  cvector_smoother( indeces )

        return self.averaged_intensity_array , self.average_slice_array, self.averaged_c_vector

    def get_id_info(self):
        return 'SD(id=%s, corrected=%s)' % (str(id(self)), str(self.baseline_corrected))

    def get_xray_data(self):
        q = self.intensity_array[0,:,0]
        data = self.intensity_array[:,:,1].T
        error = self.intensity_array[:,:,2].T
        return q, data, error

    def get_xr_data_separate_ly(self):
        D = self.intensity_array[:,:,1].T
        E = self.intensity_array[:,:,2].T
        return D, E, self.qvector, self.xray_curve

    def get_uv_data_separate_ly(self):
        D = self.conc_array
        uv_curve = self.get_uv_curve()
        return D, None, self.lvector, uv_curve

    def get_xray_scale(self):
        return self.xray_curve.max_y

    def get_cd_slice(self):
        if self.cd_slice is None:
            qmax = bisect_right(self.qvector, get_setting('cd_eval_qmax'))
            self.cd_slice = slice(0, qmax)
        return self.cd_slice

    def save_xray_data(self, out_folder):
        from molass_legacy.KekLib.NumpyUtils import np_savetxt

        ret_files = []
        for j, path in enumerate(self.datafiles):
            _, file = os.path.split(path)
            # out_path = os.path.join(out_folder, file)
            out_path = '/'.join([out_folder, file])
            print([j], out_path)
            ret_files.append(out_path)
            np_savetxt(out_path, self.intensity_array[j])

        return ret_files

    def log_id_values(self, label):
        self.logger.info("id values at %s are (%g, %g)", label, np.sum(self.intensity_array), np.sum(self.conc_array))

    def get_elution_curves(self):
        uv_curve = self.get_uv_curve()
        xr_curve = self.get_xray_curve()
        return uv_curve, xr_curve



"""
    Trimming.TrimmingUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from bisect import bisect_right
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Trimming.TrimmingInfo import TrimmingInfo

def get_default_uv_wl_trimming(sd):
    # task: unifty with the similar coding in AutoRestrictor
    uv_wl_lower_bound = get_setting("uv_wl_lower_bound")
    assert uv_wl_lower_bound is not None
    start = bisect_right(sd.lvector, uv_wl_lower_bound)
    size = len(sd.lvector)
    return TrimmingInfo(1, start, size, size)


"""
    Solvers.ABC.ConstrainedPrior.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import gamma
from importlib import reload
from pyabc import RV, DistributionBase, Parameter
import Solvers.ABC.ParameterUtils
reload(Solvers.ABC.ParameterUtils)
from Solvers.ABC.ParameterUtils import vector_to_parameter, parameter_to_vector

MAX_NUM_PARAMS = 100
PDF_FACTOR_SCALE = 1e-1     # experimental. to do this correctly, see also numpy.isclose

class ConstrainedPrior(DistributionBase):
    def __init__(self, shape, lower, upper, offset=1):
        self.num_all_params = len(lower)
        assert self.num_all_params <= MAX_NUM_PARAMS
        num_components, model_num_params = shape
        rv_list = []
        for lb, ub in zip(lower, upper):
            # rv_list.append(RV("uniform", lb, ub))
            loc = (lb + ub)/2
            scale = (ub - lb)*1/3
            rv_list.append(RV("norm", loc=loc, scale=scale))
        self.rv_list = rv_list
        self.ordering_slice = np.array(offset + np.arange(num_components) * model_num_params, dtype=int)
        n = self.num_all_params
        print("ordering_slice=", self.ordering_slice)
        self.volume_ratio = np.power(np.pi, n/2)/gamma(n/2 + 1)*np.power(2.0, -n)   # 4.13e-19 for n=37
        self.pdf_factor = 1/self.volume_ratio

    def _unconstraned_rvs(self):
        ret_values = np.zeros(len(self.rv_list))
        for i, p in enumerate(self.rv_list):
            ret_values[i] = p.rvs()
        return ret_values
    
    def rvs(self, *args, **kwargs):
        while True:
            values = self._unconstraned_rvs()
            ordering_values = values[self.ordering_slice]
            if np.min(np.diff(ordering_values)) > 0:
                return vector_to_parameter(values)

    def pdf(self, x):
        # values = np.array([x[PARAM_KEY_FORMAT % k] for k in range(self.num_all_params)])
        values = parameter_to_vector(x)
        ordering_values = values[self.ordering_slice]
        if np.min(np.diff(ordering_values)) <= 0:
            return 0.0
        # print("values=", values)
        # print("ordering_values=", ordering_values)
        # return np.product([rv.pdf(p) for rv, p in zip(self.rv_list, values)])
        p = np.product([rv.pdf(p) for rv, p in zip(self.rv_list, values)])
        # print("p=", p)
        return p * self.pdf_factor


"""

    SerialController.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import os
import glob
import threading
import logging
import psutil
import time
from openpyxl import Workbook

from molass_legacy.KekLib.BasicUtils import get_filename_extension
from .SerialDataUtils import save_xray_base_profiles
from .StageGuinier import run_gunier_analysis, make_guinier_analysis_report
from .StageExtrapolation import control_extrapolation
from .StageSummary import do_summary_stage
from molass_legacy.KekLib.ExcelCOM import cleanup_created_excels
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, save_settings
from molass_legacy.KekLib.NumpyUtils import np_savetxt
from molass_legacy.KekLib.BasicUtils import ( clear_dirs_with_retry, mkdirs_with_retry, ordinal_str )
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from .AnalysisRangeManager   import get_analysis_ranges_for_exec
from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges, report_ranges_from_analysis_ranges
from molass_legacy.KekLib.ProgressInfo import put_info, send_stop, on_stop_raise, put_error, STATE_FATAL, SAFE_FINISH
from .ProgressInfoUtil import ( NUM_SHEET_TYPES,
                                STREAM_BASECOR, STREAM_GUINIER, STREAM_ZERO_EX, STREAM_BASECOR_MAX,
                                ProgressCallback )
from DevSettings import get_dev_setting
from DriftSimulation import apply_simulated_baseline_correction
from molass_legacy.Test.Tester import ( write_to_tester_log,
                            create_log_queue, open_dev_log, write_to_dev_log, close_dev_log,
                            write_from_log_queue
                            )
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

DEBUG = False

NUM_ZX_STEPS        = ( 2 + 1 + 2 ) * 2 + 1
NUM_ZX_BOOK_STEPS   = 2

class ExecInfo:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

    def __str__( self ):
        return str( self.state )

class SerialController:
    def __init__( self, control_info):
        self.control_info = control_info
        self.cleaner = control_info.cleaner
        self.exe_thread     = None

    def start( self, detail=False ):
        serial_data = self.control_info.serial_data
        assert serial_data is not None

        serial_data.wait_until_ready()

        """
        temporarily commented out due to the following error
                pickle.dump( self.dictionary, pf )
            TypeError: cannot pickle '_tkinter.tkapp' object
        """
        # temporary fix for inconsystency bug or risk of persistent memory between threads
        save_settings()

        self.executer = SerialExecuter()
        if self.cleaner is not None:
            self.cleaner.set_cleanup(self.executer.cleanup)

        self.exe_thread = threading.Thread(
                                target=self.executer.run,
                                name='ExecutionThread',
                                args=[self.control_info],
                                )
        self.exe_thread.start()

    def is_alive(self):
        return self.exe_thread.is_alive()

    def kill( self ):
        send_stop()

    def terminate( self ):
        from molass_legacy.KekLib.ThreadUtils import terminate_thread
        self.logger = logging.getLogger( __name__ )
        self.logger.info( 'terminate' )
        if self.exe_thread is None:
            return

        self.exe_thread.join( timeout=5 )
        if self.exe_thread.is_alive():
            self.logger.warning( 'terminate_thread' )
            terminate_thread( self.exe_thread )
        cleanup_created_excels()

class SerialExecuter:
    def __init__(self):
        self.current_datafile   = None
        self.log_memory_usage   = get_dev_setting( 'log_memory_usage' )
        self.process    = psutil.Process()
        self.excel_list = []
        self.logger     = logging.getLogger( __name__ )

    def run(self, control_info):
        from molass_legacy.KekLib.ExcelCOM import CoInitialize, ExcelComClient

        self.temp_folder    = control_info.temp_folder
        self.more_multicore = control_info.more_multicore
        self.make_temp_folder()

        if self.more_multicore:
            from molass_legacy.ExcelProcess.ExcelTeller import ExcelTeller
            self.teller = ExcelTeller(log_folder=self.temp_folder)
            self.logger.info('teller created with log_folder=%s', control_info.temp_folder)

        self.parent = control_info.parent
        serial_data = control_info.serial_data
        self.stream  = STREAM_BASECOR
        scattering_correction = get_setting( 'scattering_correction' )
        start_time0 = time.time()
        if scattering_correction == 1:
            base_drift_params = get_setting( 'base_drift_params' )

            try:
                if base_drift_params is None:
                    max_progress    = len( serial_data.qvector ) - 1
                    progress_cb     = ProgressCallback( max_progress )

                    enable_xb_save  = get_dev_setting( 'enable_xb_save' )
                    return_base     = enable_xb_save == 1
                    corrected_base  = serial_data.apply_baseline_correction( control_info.mapped_info, progress_cb=progress_cb, return_base=return_base )
                    if return_base:
                        save_xray_base_profiles( serial_data, corrected_base )
                else:
                    max_progress    = len( serial_data.datafiles ) - 1
                    progress_cb     = ProgressCallback( max_progress )

                    apply_simulated_baseline_correction( base_drift_params, serial_data.intensity_array, progress_cb=progress_cb )

                self.logger.info( 'Baseline correction (degree=%d) applied to scattering data.' % ( get_setting('baseline_degree') ) )
            except ( RuntimeError ) as exc:
                self.logger.warning( exc )
            except:
                etb = ExceptionTracebacker()
                self.logger.error( str(etb) )
                put_error( self.stream, error_state=STATE_FATAL )
                return

            progress_cb( max_progress )

        start_time1 = time.time()
        self.seconds_correction = int(start_time1 - start_time0)
        # self.input_smoothing = get_setting( 'input_smoothing' )
        self.input_smoothing = 1

        self.temp_books     = []
        self.temp_books_atsas   = []
        self.data_folder    = control_info.data_folder
        self.conc_folder    = control_info.conc_folder
        self.work_folder    = control_info.work_folder
        self.maintenance_log    = control_info.maintenance_log
        self.maintenance_mode   = control_info.maintenance_log is not None
        self.env_info       = control_info.env_info
        self.excel_is_available = self.env_info.excel_is_available
        self.atsas_is_available = self.env_info.atsas_is_available
        self.guinier_folder = control_info.guinier_folder
        d_, f_ = os.path.split( control_info.serial_file )
        self.guinier_file   = os.path.join( self.guinier_folder, f_ )
        self.stamp_file     = control_info.stamp_file
        self.quality_file   = os.path.join( self.guinier_folder, '--quality_array.csv' )
        self.serial_file    = control_info.serial_file
        self.book_file      = control_info.book_file
        self.min_analysis_value    = control_info.min_analysis_value
        self.range_type     = control_info.range_type
        self.zx             = control_info.zx
        self.mapped_info    = control_info.mapped_info
        self.known_info_list = control_info.known_info_list
        self.zx_out_folder  = None
        self.use_simpleguinier  = get_dev_setting( 'use_simpleguinier' )
        if self.use_simpleguinier == 1:
            current_quality_weighting = get_setting( 'quality_weighting' )
            new_weighting = [
                                0.2,    # basic_quality
                                0.2,    # positive_score
                                0.2,    # fwd_consistency
                                0.2,    # rg_stdev_score
                                0.2,    # q_rg_score
                              ]
            set_setting( 'quality_weighting', new_weighting )

        self.stream  = STREAM_GUINIER
        self.zx_summary_list = []
        self.zx_summary_list2 = []

        try:
            if self.maintenance_mode:
                self.logger.warning( 'executing in the maintenance mode' )

            if self.log_memory_usage == 1:
                m_info = self.process.memory_info()
                memory_usage    = " " + str( ( m_info.vms, ) )
            else:
                memory_usage    = ""

            self.preview_params = control_info.preview_params
            pdata = self.preview_params[0]
            self.doing_sec = pdata.is_for_sec

            if self.doing_sec:
                self.xr_j0 = serial_data.xr_j0
            else:
                self.xr_j0 = pdata.sd.xr_j0

            analysis_ranges = control_info.analysis_ranges
            range_type = control_info.range_type
            data_folder = control_info.data_folder
            conc_folder = control_info.conc_folder
            book_file = control_info.book_file

            if analysis_ranges is None:
                # set_analysis_range must be delayed for the range_type '3'
                self.report_ranges = None
            else:
                self.set_analysis_range( range_type, analysis_ranges )
                self.report_ranges = report_ranges_from_analysis_ranges(self.xr_j0, self.applied_ranges)

            self.logger.info( 'Xray scattering folder=' + data_folder )
            self.logger.info( 'UV absorbance folder=' + conc_folder )
            self.logger.info( 'concetration factor has been set to %g' % serial_data.conc_factor )
            self.logger.info( 'range_type=%d and analysis_ranges=%s%s' % ( range_type, self.report_ranges, memory_usage ) )

            write_to_tester_log( 'book_file=' + book_file + '\n' )
            if self.maintenance_mode:
                create_log_queue()
                open_dev_log( self.maintenance_log )
                write_to_dev_log( 'Doing ' + data_folder + '\n' )

            self.save_settings()

            self.set_control_data( serial_data )

            write_to_tester_log( 'guinier start\n' )

            run_gunier_analysis( self )

            if analysis_ranges is None:
                self.set_analysis_range( range_type, analysis_ranges )
            else:
                pass
                # already set at the start (above)

            if self.excel_is_available:
                if self.more_multicore:
                    self.excel_client = None
                else:
                    CoInitialize()
                    self.excel_client = ExcelComClient()
                self.result_wb = None
            else:
                self.excel_client = None
                self.result_wb = Workbook()
                # self.result_wb.remove_sheet(self.result_wb.active)

            make_guinier_analysis_report( self )

            write_to_tester_log( 'guinier finish\n' )
            start_time2 = time.time()
            self.seconds_guinier = int(start_time2 - start_time1)

            self.stream  = STREAM_ZERO_EX
            if self.zx:
                control_extrapolation( self )

            start_time3 = time.time()
            self.seconds_extrapolation = int(start_time3 - start_time2)

            do_summary_stage( self )

            start_time4 = time.time()
            self.seconds_summary = int(start_time4 - start_time3)

            self.cleanup()

            if self.log_memory_usage == 1:
                m_info = self.process.memory_info()
                memory_usage    = " " + str( ( m_info.vms, ) )
            else:
                memory_usage    = ""

            self.logger.info( 'Finished.' + memory_usage )
            self.logger.info( "The report has been saved to '%s'" % ( book_file ) )

            put_info( (STREAM_ZERO_EX,2000), SAFE_FINISH )

            put_info( (STREAM_ZERO_EX,2000), 1 )

            if self.maintenance_mode:
                write_from_log_queue()
                write_to_dev_log( 'Finished ' + data_folder + '\n' )
                close_dev_log()

        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.error( etb )
            if self.current_datafile is not None:
                self.logger.error( 'current_datafile:' + str(self.current_datafile) )   # str to allow None
            self.error_cleanup()
            self.error_wrapup( book_file )
            put_error( self.stream, error_state=STATE_FATAL )

        if self.use_simpleguinier == 1:
            set_setting( 'quality_weighting', current_quality_weighting )

    def stop_check( self ):
        def log_closure(cmd):
            # this closure is expected to be called only in cancel operations
            self.logger.info("cmd=%s", str(cmd))
        on_stop_raise(cleanup=self.error_cleanup, log_closure=log_closure)

    def make_temp_folder( self ):
        try:
            clear_dirs_with_retry( [self.temp_folder] )
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.error( etb )
            raise exc

    def compute_num_guinier_steps( self, num_datafiles ):
        return 1 + num_datafiles + 5

    def set_control_data( self, serial_data ):
        self.serial_data = serial_data
        self.datafiles = serial_data.datafiles 
        self.num_datafiles = len( serial_data.datafiles )
        self.num_peaks_to_exec = len( self.mapped_info.x_ranges )

    def get_intensity_array( self ):
        """
            input_smoothing
                1 : 
                2 : 
        """

        self.serial_data.apply_data_reduction()
        # if get_dev_setting( 'no_usable_q_limit' ) == 1:
        if True:
            self.usable_slice = slice( 0, self.serial_data.intensity_array.shape[1] )
        else:
            self.usable_slice = self.serial_data.get_usable_slice()
            if self.usable_slice.stop is None:
                predicate = ' is not set'
            else:
                predicate = ' is set at ' + str(self.usable_slice.stop)
            self.logger.info( 'Usable Xray scattering data limit' + predicate )

        """
            concentration vector variation

            SerialData          oc_vector
                                mc_evector

            SerialControler     c_vector  ( mc_vector or smoothed mc_vector )

            ZeroExtrapolator    c_vector
        """

        self.using_avareged_files = False
        if self.input_smoothing == 1:
            num_curves_averaged = get_setting( 'num_curves_averaged' )
            intensity_array_, average_slice_array, c_vector = self.serial_data.get_averaged_data( num_curves_averaged )
            assert c_vector is not None

            # save
            save_averaged_data = get_setting( 'save_averaged_data' )
            if save_averaged_data == 1:
                self.using_avareged_files = True        # TODO: check consistency
                self.save_smoothed_data( intensity_array_, average_slice_array )
        else:
            intensity_array_ = self.serial_data.intensity_array
            c_vector = self.serial_data.mc_vector

        if self.doing_sec:
            self.c_vector = c_vector
        else:
            # temp fix; TODO: do this correctly
            pdata = self.preview_params[0]
            self.c_vector = pdata.mc_vector

        return intensity_array_

    def save_smoothed_data( self, intensity_array, average_slice_array ):
        self.averaged_data_folder = get_setting( 'averaged_data_folder' )
        if not os.path.exists( self.averaged_data_folder ):
            mkdirs_with_retry( self.averaged_data_folder )

        avg_file_postfix = get_setting( 'avg_file_postfix' )
        extension = '.' + get_filename_extension( self.datafiles[0] )
        self.averaged_datafiles = []
        for i, intensity in enumerate( intensity_array ):
            _, filename = os.path.split( self.datafiles[i] )
            # print( 'saving', filename )
            filepath = '/'.join( [ self.averaged_data_folder, filename.replace( extension, avg_file_postfix + extension ) ] )
            with open( filepath, "wb" ) as fh:
                slice_ = average_slice_array[i]
                fh.write( str.encode( '# Created by averaging the following %d files with []-numbering starting from 1.\n' % ( slice_.stop - slice_.start ) ) )
                for j in range( slice_.start, slice_.stop ):
                    fh.write( str.encode( '# [%d] %s\n' % ( j+1, self.datafiles[j] ) ) )
                fh.write( str.encode( "#\n# Q\tIntensity\tError\n" ) )
                fh.close()
            np_savetxt( filepath, intensity, mode="a" )
            self.averaged_datafiles.append( filepath )

    def set_analysis_range( self, range_type, analysis_ranges ):
        ranges = get_analysis_ranges_for_exec( self, range_type, analysis_ranges, self.logger )
        # self.logger.info( arg_join( 'set_analysis_range: ranges=', ranges ) )

        # change ranges to internal ranges
        self.applied_ranges = ranges

    def write_state( self, state ):
        self.clear_state()
        path = self.temp_folder + '/--state-' + state
        fh = open( path, "w" )
        fh.close()

    def clear_state( self ):
        state_files = glob.glob( self.temp_folder + '/--state*' )
        for path in state_files:
            os.remove( path )

    def error_wrapup( self, book_file ):
        self.logger.warning( 'Error wrap-up start.' )
        try:
            self.merge_books()
            self.cleanup()
            self.logger.warning( 'Error wrap-up done.' )
            if len(self.temp_books) > 0:
                self.logger.warning( "The incomplete report has been saved to '%s'" % ( book_file ) )
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.error( etb )
            self.logger.info( "Temporary files are left as worked for investigation." )

    def save_settings( self ):
        save_folder = self.temp_folder.replace( '.temp', '.save' )
        clear_dirs_with_retry( [ save_folder ] )
        save_path = save_folder + '/serial_settings.dump'
        save_settings( file=save_path )

    def cleanup(self):
        self.logger.info("Cleanup started. This may take some time (not more than a few minutes). Please be patient.")

        if self.more_multicore:
            self.teller.stop()   # must be done before the removal below of the temp books

        if self.excel_is_available:
            if self.more_multicore:
                pass
            else:
                from molass_legacy.KekLib.ExcelCOM import CoUninitialize
                self.excel_client.quit()
                self.excel_client = None
                CoUninitialize()

            for path in self.temp_books + self.temp_books_atsas:
                os.remove( path )

        self.logger.info("Cleanup done.")

    def error_cleanup(self):
        self.cleanup()
        cleanup_created_excels()



"""
    SecTheory.DiporeSimulation.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec
from scipy.interpolate import UnivariateSpline

def demo():
    import molass_legacy.KekLib.DebugPlot as plt
    from SecTheory.SecCF import gec_dipore_phi
    from SecTheory.SecPDF import FftInvPdf
    from SecTheory.ColumnIllust import plot_column_illust, plot_column_legend

    dipore_pdf = FftInvPdf(gec_dipore_phi)

    x = np.arange(200)

    me = 0.1
    mp = 1.5
    rp0 = 76
    rp1 = 90
    rp2 = 40
    rg = 35
    N = 171
    T = 0.63
    t0 = 20

    rg_array = np.array([70, 65, 60, 55, 50, 45, 40, 35, 30, 25, 20, 15, 10, 5, 3, 1])
    rho0 = rg_array/rp0
    rho0[rho0 > 1] = 1
    np0 = N*(1 - rho0)**me
    tp0 = T*(1 - rho0)**mp

    rho1 = rg_array/rp1
    rho1[rho1 > 1] = 1
    np1 = N*(1 - rho1)**me
    tp1 = T*(1 - rho1)**mp

    rho2 = rg_array/rp2
    rho2[rho2 > 1] = 1
    np2 = N*(1 - rho2)**me
    tp2 = T*(1 - rho2)**mp

    y1_list = []
    confpoints1 = []
    y2_list = []
    confpoints2 = []
    k = 0
    for np0_, tp0_, np1_, tp1_, np2_, tp2_ in zip(np0, tp0, np1, tp1, np2, tp2):
        y1 = dipore_pdf(x, 0.5*np0_, tp0_, 0.5*np0_, tp0_, t0)
        y1_list.append(y1)
        j = np.argmax(y1)
        confpoints1.append((x[j], rg_array[k]))

        y2 = dipore_pdf(x, 0.5*np1_, tp1_, 0.5*np2_, tp2_, t0)
        y2_list.append(y2)
        j = np.argmax(y2)
        confpoints2.append((x[j], rg_array[k]))
        k += 1

    confpoints1 = np.array(confpoints1)
    confpoints2 = np.array(confpoints2)

    def create_spline(confpoints):
        return UnivariateSpline(confpoints[:,0], confpoints[:,1], s=0)

    spline1 = create_spline(confpoints1)
    spline2 = create_spline(confpoints2)

    x_ = x[t0+8:130]

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(13.5, 9))
        ax1, ax2, ax3 = axes[0,:]
        fig.suptitle("Comparison of Conformance Curves between Monopore and Dipore Columns", fontsize=20)
        ax1.set_title("Monopore Column with Poresize=%g" % rp0, fontsize=16)
        ax2.set_title("Dipore Column with Poresizes=%g, %g" % (rp1, rp2), fontsize=16)
        ax3.set_title("Overlay of the Conformance Curves", fontsize=16)

        ylimits = []
        for ax, y_list, spline, confpoints, color1, color2, name in [
                    (ax1, y1_list, spline1, confpoints1, "yellow", "red", "Monopore"),
                    (ax2, y2_list, spline2, confpoints2, "cyan", "green", "Dipore")]:

            axt = ax.twinx()
            axt.grid(False)

            for y, rg_ in zip(y_list, rg_array):
                ax.plot(x, y, alpha=0.5, label="$R_g=%.3g$" % rg_)

            for k, ax_ in enumerate([axt, ax3]):
                label = None if k == 0 else name
                ax_.plot(x_, spline(x_), color=color1, lw=3, alpha=0.5, label=name)
                ax_.plot(*confpoints.T, "o", color=color2, alpha=0.5)

            # ax.legend()
            ylimits.append(ax.get_ylim())

        ax3.legend()
        ax3.set_xlim(0, 200)

        ylimits = np.array(ylimits)
        ymin = np.min(ylimits[:,0])
        ymax = np.max(ylimits[:,1])
        for ax in ax1, ax2:
            ax.set_ylim(ymin, ymax)

        plot_column_illust(axes[1,0], [rp0], ["white"])
        plot_column_illust(axes[1,1], [rp1, rp2], ["orange", "yellow"])
        plot_column_legend(axes[1,2], [rp1, rp0, rp2], ["orange", "white", "yellow"])

        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



"""
   BaselineParams.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.SerialSettings import get_setting

def get_num_baseparams():
    unified_baseline_type = get_setting("unified_baseline_type")
    if unified_baseline_type == 1:
        return 2
    elif unified_baseline_type == 2:
        return 3
    elif unified_baseline_type == 3:
        return 3
    else:
        assert False



"""
    Models/RateTheory/stlcGRM.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from stlc import grm

cm = 1e-2
minute = 60
dt = 1
qm = 1.
nr = 10
nz = 20

def grm_impl(x, zl, rp, ec, ep, kf, Dax, Dp, Ds, u, ka, kd, tinj, cinj):

    # tmax = len(x) * minute  #[min]
    tmax = len(x)

    def step(t: float, tinj: float) -> float:
        return float(t <= tinj)

    parameters0 = grm.ModelParameters(c0=cinj,
                                          Dax=Dax,
                                          Dp=Dp,
                                          Ds=Ds,
                                          ka=ka,
                                          kd=kd,
                                          kf=kf,
                                          qm=qm,
                                          ip=lambda t: step(t, tinj))

    model = grm.GeneralRateModel(u=u,
                                         ep=ep,
                                         ec=ec,
                                         zl=zl,
                                         rp=rp,
                                         nz=nz,
                                         nr=nr,
                                         component_parameters=[parameters0])
    sol = grm.solve(model, tmax, dt)

    y = sol.c[0, -1, :]
    return y

def guess(x, y):

    zl = 1.7 * cm  #[cm]
    rp = 0.004 * cm  #[cm]
    ec = 0.4
    ep = 0.333
    kf = 0.01 * cm  #[cm]
    Dax = 0.002 * cm**2 / minute  #[cm^2 min^-1]
    Dp = 3.003 * 1e-6 * cm**2 / minute  #[cm^2 min^-1]
    Ds = 0.  #[cm^2 min^-1]
    u = 0.6 * cm / minute  #[cm min^-1]
    ka = 2.5  #[s^-1]
    kd = 1  #[s^-1]
    tinj = 50 * minute  #[min]
    cinj = 1  #[mol /m^-3]

    init_params = (zl, rp, ec, ep, kf, Dax, Dp, Ds, u, ka, kd, tinj, cinj)

    def objective(p):
        y_ = grm_impl(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)

    print("ret.x=", ret.x)

    """
    method="Nelder-Mead"
    ret.x= [ 4.09867334e-07  5.17007792e-05  1.69859416e-06  3.39363091e-01
  8.00288751e-05  2.96235961e-09  5.08620741e-12 -8.87774587e-05
  2.14400226e-04  1.07395544e+01  1.09575930e+00  2.03315646e+02
  1.27581541e+00]
    """

    return ret.x



"""
    Solvers.UltraNesst.CallbacktxtPlot.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.FvScoreConverter import convert_score

def plot_callback_txt_impl(caller):
    from importlib import reload
    import Optimizer.FuncReloadUtils
    reload(Optimizer.FuncReloadUtils)
    from molass_legacy.Optimizer.FuncReloadUtils import reload_optimizer
    print("plot_callback_txt_impl")
    caller.update_information()
    canvas = caller.canvas
    optimizer = canvas.fullopt
    fv_array, x_array = canvas.demo_info[0:2]
    index = 2
    temp_params = x_array[index]
    fv = optimizer.objective_func(temp_params)
    sv = convert_score(fv)
    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
        fig.suptitle("plot_callback_txt_impl: index=%d" % index, fontsize=20)
        axt = ax2.twinx()
        axt.grid(False)
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objective Function Scores in SV=%.3g" % sv, fontsize=16)
        optimizer.objective_func(temp_params, plot=True, axis_info=(fig, (ax1, ax2, ax3, axt)))
        fig.tight_layout()
        plt.show()


# coding: utf-8
"""

    PlotAnnotation.py

        recognition of peaks

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""

def add_flow_change_annotation( ax, flow_changes, mapper ):
    xmin1, xmax1 = ax.get_xlim()
    ymin1, ymax1 = ax.get_ylim()
    yoffset = ( ymax1 - ymin1 )*0.2
    for i, fc in enumerate( flow_changes ):
        if fc is None:
            continue
        xoffset = ( xmax1 - xmin1 )*( 0.05 if fc < 100 else 0 )
        y_ = mapper.a_spline( fc )
        ax.annotate( "flow change", xy=(fc, y_),
                        xytext=( fc + xoffset, y_ + yoffset ),
                        ha='center',
                        arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05 ),
                        )



# coding: utf-8
"""

    TkAppInit.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

    not yet used as of 20200630

"""

from MultiMonitor import get_selected_monitor

class TkAppInit:
    def __init__(self, root):
        m = get_selected_monitor()
        print(m)
        if m.width >= 2560:
            root.tk.call('tk', 'scaling', '-displayof', '.', 1)



"""
    Models.Stochastic.LognormalGuess.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.integrate import quad
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.BasicModels import robust_single_pore_pdf
import Models.Stochastic.LognormalPoreFunc
reload(Models.Stochastic.LognormalPoreFunc)
from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func, lognormal_pore_func
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma, compute_mode, compute_stdev
from molass_legacy.Models.ModelUtils import compute_raw_moments1, compute_area_props

def guess_lognormalpore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, monopore_params, progress_cb_info=None, debug=False):
    from molass_legacy.Models.Stochastic.SecModelUtils import SIGMA_LOWER_BOUND
    import Simulative.LognormalPsd
    reload(Simulative.LognormalPsd)
    from Simulative.LognormalPsd import lognormalpore_model_interactive_impl

    # debug = True
    # print("monopore_params=", monopore_params, "peak_rgs=", peak_rgs)
    N, T, x0, me, mp, poresize = monopore_params[0:6]
    # init_sigma = DEFAULT_SIGMA
    # init_mu = np.log(poresize)
    init_mu, init_sigma = compute_mu_sigma(poresize, poresize*0.2)
    print("init_mu, init_sigma=", init_mu, init_sigma)
    init_sigma = max(init_sigma, SIGMA_LOWER_BOUND)
    print("init_mu, init_sigma=", init_mu, init_sigma)
    if debug:
        scales = monopore_params[6:]
        print("poresize=", poresize, compute_mode(init_mu, init_sigma))
        print("stdev=", compute_stdev(init_mu, init_sigma))
        with plt.Dp(button_spec=["OK", "Cancel"]):
            fig, ax = plt.subplots()
            ax.set_title("Lognormalpore vs. Monopore")
            ax.plot(x, y)
            lnp_cy_list = []
            mnp_cy_list = []
            for k, (rg, scale) in enumerate(zip(peak_rgs, scales)):
                color = "C%d" % (k+1)
                cy = lognormal_pore_func(x, scale, N, T, me, mp, init_mu, init_sigma, rg, x0)
                lnp_cy_list.append(cy)
                ax.plot(x, cy, alpha=0.3, color=color)
                rho = min(1, rg/poresize)
                ni_ = N * (1 - rho)**me
                ti_ = T * (1 - rho)**mp
                cy = scale*robust_single_pore_pdf(x - x0, ni_, ti_)
                mnp_cy_list.append(cy)
                ax.plot(x, cy, ":", color=color)
            lnp_ty = np.sum(lnp_cy_list, axis=0)
            mnp_ty = np.sum(mnp_cy_list, axis=0)
            ax.plot(x, lnp_ty, "-r", label="Lnpore")
            ax.plot(x, mnp_ty, "-b", label="Monopore")
            ax.legend()

            fig.tight_layout()
            ret = plt.show()
        if not ret:
            return

    abort = False
    use_multi_factor_fv = False
    if use_multi_factor_fv:
        raw_moments = np.array([M[0] for M in egh_moments_list])

    def mu_sigma_scales_objective(p, title=None):
        nonlocal abort
        x0_, mu, sigma = p[0:3]
        cy_list = []
        for rg, scale in zip(peak_rgs, p[3:]):
                cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0_)
                cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        fv = np.sum((ty - y)**2)
        if use_multi_factor_fv:
            raw_moments_ = compute_raw_moments1(x, cy_list)
            props_ = compute_area_props(cy_list)
            fv = fv**2 + 0.0001*np.sum((raw_moments_ - raw_moments)**2)**2 + 0.0001*np.sum((props_ - props)**2)**2

        if False:
            title = "minimize iteration"
        if title is not None:
            print("%s fv=%.3g" % (title, fv))
            print("%s p=%s" % (title, repr(p)))
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                ret = plt.show()
            if not ret:
                abort = True
        if np.isnan(fv):
            fv = np.inf
        return fv
    
    num_components = len(peak_rgs)
    bounds = [(x0-50, x0+50), (0, None), (SIGMA_LOWER_BOUND, None)] + [(0, None)]*num_components
    init_scales = monopore_params[6:]
    init_params = np.concatenate([[x0, init_mu, init_sigma], init_scales])
    if debug:
        print("init_params =", init_params)
        mu_sigma_scales_objective(init_params, title="init debug")
        if abort:
            return
        demo_params = np.concatenate([[N, T, x0, me, mp], init_params[1:]])
        ret = lognormalpore_model_interactive_impl(x, y, demo_params, peak_rgs, title="guess_lognormalpore_params_using_moments debug")
        if not ret:
            return
        mu_sigma_scales_objective(init_params, title="before LNP minimize")
        if abort:
            return
    if progress_cb_info is None:
        callback = None
    else:
        minp, maxp, progress_cb = progress_cb_info
        max_iters = len(init_params) * 200
        call_counter = 0
        def callback_func(x, *args):
            nonlocal call_counter
            call_counter += 1
            if call_counter % 10 == 0:
                progress_cb(minp + (maxp-minp)*call_counter/max_iters)
        callback = callback_func
    res = minimize(mu_sigma_scales_objective, init_params, bounds=bounds, method="Nelder-Mead", callback=callback)
    if debug:
        print("res.x =", res.x)
        mu_sigma_scales_objective(res.x, title="after LNP minimize")
    x0, mu, sigma = res.x[0:3]
    opt_params = np.concatenate([[N, T, x0, me, mp, mu, sigma], res.x[3:]]) 
    return opt_params


"""
    Processes.Loader.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""

def loader(si, pn, queues):
    from time import sleep
    from .SharedInfo import QUEUE_GUI, QUEUE_LOADER, QUEUE_RG, QUEUE_RECOG
    from LoaderProcess.DataLoader import DataLoader
    print("loader", si)

    gui_queue = queues[QUEUE_GUI]
    loader_queue = queues[QUEUE_LOADER]
    rg_queue = queues[QUEUE_RG]
    recog_queue = queues[QUEUE_RECOG]

    dl = DataLoader(si, rg_queue)

    while True:
        in_folder = loader_queue.get()
        if in_folder == "__stop__":
            break

        print(f'Working on {in_folder}')
        dl.load(in_folder)
        gui_queue.put(1)
        recog_queue.put(1)
        print(f'Finished {in_folder}')

    print("loader stop")



"""
    PairedDataSets.py.

    Copyright (c) 2020, 2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from .ProbDensityUtils import generate_sample_data
from .GroupingMatrix import num_to_char

def get_cut_num_peaks(ecurve):
    """
    be aware that it is assumed here
    that this ecurve is not yet cut,
    i.e. ecurve.x retains the original elution numbers.
    """
    if ecurve is None:
        # as in 20190309_3
        num_peaks = 1
    else:
        peaks = ecurve.get_major_peak_info(add_j0=True)
        qmm_window_slice = get_setting('qmm_window_slice')
        print('qmm_window_slice=', qmm_window_slice, 'peaks=', peaks)
        if qmm_window_slice is None:
            num_peaks = len(peaks)
        else:
            num_peaks = 0
            start, stop = qmm_window_slice.start, qmm_window_slice.stop
            for rec in peaks:
                top_x = rec[1]
                if top_x >= start and top_x < stop:
                    num_peaks += 1
    return num_peaks

def get_denoise_rank_impl(ecurve, factor=None):
    num_peaks = get_cut_num_peaks(ecurve)
    if factor is None:
        factor = get_setting('conc_dependence')
    return min(16, num_peaks*factor + (num_peaks-1) + 2)

def get_num_components(ecurve):
    num_peaks = get_cut_num_peaks(ecurve)
    print('num_peaks=', num_peaks)
    return min(16, 8 + num_peaks*2)

WAVELENGTH_POINTS = 280, 260
ANGLE_POINTS = 0.02, 0.04

class E11nInfo:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

class PairedDataSets:
    def __init__(self, in_folder, kwargs):
        self.logger = logging.getLogger(__name__)
        lpm_correct = kwargs.pop('lpm_correct', False)

        if lpm_correct:
            from molass_legacy.DataStructure.CorrectedData import CorrectedXray, CorrectedUv
            rx = CorrectedXray(in_folder)
            try:
                ru = CorrectedUv(in_folder)
            except AssertionError:
                ru = rx.make_uv_proxy()
        else:
            from molass_legacy.DataStructure.RawData import RawXray, RawUv
            rx = RawXray(in_folder)
            try:
                ru = RawUv(in_folder)
            except AssertionError:
                ru = rx.make_uv_proxy()

        self.kwargs = kwargs
        self.pair = (ru, rx)

    def get_i_pos_values(self, dtype):
        values = WAVELENGTH_POINTS if dtype == 'UV' else ANGLE_POINTS
        return values

    def generate_sample_datasets(self, quad=False, bubble_care=False):
        kwargs = self.kwargs
        ru, rx = self.pair

        lpm_2d = kwargs.pop('lpm_2d', False)
        smoothing = kwargs.get('smoothing', False)

        i = ru.get_row_index(WAVELENGTH_POINTS[0])
        y1 = ru.data[i,:]
        if lpm_2d:
            from molass_legacy.DataStructure.LPM import get_corrected
            y1 = get_corrected(y1)

        i = rx.get_row_index(ANGLE_POINTS[0])
        y2 = rx.data[i,:]
        if bubble_care:
            raw_xr = rx.xr
            raw_y2 = raw_xr.data[i,:]
            from molass_legacy.SerialAnalyzer.AbnormalityCheck import bubble_check_impl
            exclude = bubble_check_impl(raw_y2)
            print('generate_sample_datasets: exclude=', exclude)
            if len(exclude) > 0:
                raw_xr.exclude_elution(exclude)
                self.logger.warning("raw elution %s have been excluded and interpolated.", str(exclude))
                rx.initialize(raw_xr)
                y2 = rx.data[i,:]

        if lpm_2d:
            from molass_legacy.DataStructure.LPM import get_corrected
            y2 = get_corrected(y2)

        if smoothing:
            from molass_legacy.KekLib.SciPyCookbook import smooth
            y1 = smooth(y1)
            y2 = smooth(y2)

        data1 = generate_sample_data(y1, 2)
        data2 = generate_sample_data(y2, 2)

        if not quad:
            return y1, y2, data1, data2

        i = ru.get_row_index(WAVELENGTH_POINTS[1])
        y3 = ru.data[i,:]
        if lpm_2d:
            from molass_legacy.DataStructure.LPM import get_corrected
            y3 = get_corrected(y3)

        i = rx.get_row_index(ANGLE_POINTS[1])
        y4 = rx.data[i,:]
        if lpm_2d:
            from molass_legacy.DataStructure.LPM import get_corrected
            y4 = get_corrected(y4)

        if smoothing:
            from molass_legacy.KekLib.SciPyCookbook import smooth
            y3 = smooth(y3)
            y4 = smooth(y4)

        data3 = generate_sample_data(y3, 2)
        data4 = generate_sample_data(y4, 2)

        return [y1, y3, y2, y4], [data1, data3, data2, data4]

    def guess_denoise_rank(self, ipe):
        forced_denoise_rank = get_setting('forced_denoise_rank')
        if forced_denoise_rank is None:
            try:
                denoise_rank = get_denoise_rank_impl(self.pair[1].ecurve)
            except:
                denoise_rank = 10
        else:
            denoise_rank = forced_denoise_rank
        self.logger.info("denoise_rank is assumed to be %d", denoise_rank)
        return denoise_rank

    def guess_denoise_rank_try(self, ipe):
        return 21 if ipe else 11

    def draw_exprapolated(self, ano, ax, argC, size=None, ipe=False):
        ux = 0 if ano < 2 else 1

        if size is None:
            size = argC.shape[0]

        if ux == 0:
            x = self.pair[ux].wvector
            ax.set_xlim(200, 300)
            C = argC
        else:
            x = self.pair[ux].qvector
            ax.set_xlim(0, 0.25)
            if ipe:
                C = np.vstack([argC, argC**2])
            else:
                C = argC

        rank = self.guess_denoise_rank(ipe)
        M = get_denoised_data(self.pair[ux].data, rank=rank)
        Cinv = np.linalg.pinv(C)
        P = np.dot(M, Cinv)

        for k in range(2):
            ax.plot(-10, 0, 'o')

        lines = []
        ylist = []
        for k, py in enumerate(P.T[:size]):
            y_ = py if ux == 0 else np.log10(py)
            line, = ax.plot(x, y_, label='c-%s' % num_to_char(k))
            lines.append(line)
            ylist.append(y_)
        ax.legend(loc='upper right')
        v = self.pair[ux].vector
        E = self.pair[ux].error
        return lines, np.array(ylist), E11nInfo(M=M, C=C, Cinv=Cinv, P=P, E=E, v=v, size=size)

    def update_exprapolated(self, ano, lines, C, ipe=False):
        ux = 0 if ano < 2 else 1
        rank = self.guess_denoise_rank(ipe)
        M = get_denoised_data(self.pair[ux].data, rank=rank)
        Cinv = np.linalg.pinv(C)
        P = np.dot(M, Cinv)
        for k, py in enumerate(P.T):
            y_ = py if ux == 0 else np.log10(py)
            lines[k].set_ydata(y_)



"""
    RgProcess.RgCurveUtils.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
from .RgCurve import check_rg_folder, RgCurve

rg_root_folder = None

def make_and_save_rgcurves(in_root_folder, out_root_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.SerialAnalyzer.DataUtils import serial_folder_walk
    global rg_root_folder

    for path in [in_root_folder, out_root_folder]:
        assert os.path.exists(path)

    rg_root_folder = out_root_folder

    set_setting("test_pattern", 0)
    print(in_root_folder)
    serial_folder_walk(in_root_folder, do_a_folder)

def rg_folder_from_in_folder(in_folder):
    global rg_root_folder

    if rg_root_folder is None:
        from molass_legacy.SerialAnalyzer.DataUtils import get_pytools_folder
        pytools = get_pytools_folder()
        rg_root_folder = os.path.join(pytools, "RgCurves")

    nodes = in_folder.replace("/", "\\").split("\\")
    rg_folder = os.path.join(rg_root_folder, "\\".join(nodes[3:]))
    return rg_folder

def do_a_folder(in_folder, uv_folder, plot):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from SecSaxs.DataTreatment import DataTreatment

    print(in_folder)
    if False:
        if in_folder < r"E:\PyTools\Data/20191001":
        # if in_folder < r"E:\PyTools\Data/20211001":
            return True, None

    rg_folder = rg_folder_from_in_folder(in_folder)
    print(rg_folder)
    ok = check_rg_folder(rg_folder)
    if ok:
        return True, None

    if os.path.exists(rg_folder):
        from shutil import rmtree
        rmtree(rg_folder)
    os.makedirs(rg_folder)

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=1, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()
    rg_curve = RgCurve(qv, ecurve, D, E)
    rg_curve.export(rg_folder)

    return True, None

def make_an_rg_folder(in_folder, rg_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from SecSaxs.DataTreatment import DataTreatment
    from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings

    clear_temporary_settings()

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=1, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()
    rg_curve = RgCurve(qv, ecurve, D, E)
    rg_curve.export(rg_folder)

def make_rg_folders(in_folder_list, out_root_folder):
    for k, in_folder in enumerate(in_folder_list):
        rg_folder = os.path.join(out_root_folder, '%03d' % k)
        os.makedirs(rg_folder)
        make_an_rg_folder(in_folder, rg_folder)

def plot_rg_curve(ax, rg_curve, color='gray', with_qualities=False, quality_scale=1, label=None):
    if label is None:
        label = "SAXS Rg Curve"
    for k, (x, y, rg) in enumerate(rg_curve.get_curve_segments()):
        label_ = label if k == 0 else None     # to avoid multiple appearances in the legend 
        ax.plot(x, rg, color=color, alpha=0.5, label=label_)
        if with_qualities:
            qu = rg_curve.qualities[k]
            label_ = "SAXS Rg Quality Curve" if k == 0 else None # to avoid multiple appearances in the legend 
            ax.plot(x, qu*quality_scale, color='green', alpha=0.5, label=label_)


# coding: utf-8
"""

    ファイル名：    SimulatedAnnealing.py

    処理内容：      Simulated Annealing

    original:
        http://apmonitor.com/me575/index.php/Main/SimulatedAnnealing

    animation, etc. by Masatsuyo Takahashi,

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
## Generate a contour plot
# Import some other libraries that we'll need
# matplotlib and numpy packages must also be installed
import matplotlib
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec    import GridSpec
import matplotlib.animation as animation

import random

##################################################
# Simulated Annealing
##################################################
class SimulatedAnnealing:
    def __init__( self ):
        pass

    def minimize( self, objective_f, xrange=None, start=None, seed=None, xconstaints=None, n=50, m=50 ):

        assert( xrange is not None )
        xrange = np.array( xrange )

        # Start location
        if start is None:
            x_start = np.average( xrange, axis=1 )
        else:
            x_start = start

        # Number of cycles
        # Number of trials per cycle
        # Number of accepted solutions
        na = 0.0
        # Probability of accepting worse solution at the start
        p1 = 0.7
        # Probability of accepting worse solution at the end
        p50 = 0.001
        # Initial temperature
        t1 = -1.0/np.log(p1)
        # Final temperature
        t50 = -1.0/np.log(p50)
        # Fractional reduction every cycle
        frac = (t50/t1)**(1.0/(n-1.0))
        # Initialize x
        x = np.zeros((n+1,2))
        x[0] = x_start
        xi = np.zeros(2)
        # xi = x_start
        na = na + 1.0
        # Current best results so far
        xc = np.zeros(2)
        xc = x[0]
        fc = objective_f(x_start)
        fs = np.zeros(n+1)
        fs[0] = fc

        # Current temperature
        t = t1
        ts = np.zeros(n+1)
        ts[0] = t1

        # DeltaE Average
        DeltaE_avg = 0.0

        if seed is not None:
            random.seed( seed )

        for i in range(n):
            print('Cycle: ' + str(i) + ' with Temperature: ' + str(t))
            for j in range(m):

                while True:
                    # Generate new trial points
                    xi[0] = xc[0] + random.random() - 0.5
                    xi[1] = xc[1] + random.random() - 0.5
                    # Clip to upper and lower bounds
                    xi[0] = max( min(xi[0], xrange[0,1]), xrange[0,0])
                    xi[1] = max( min(xi[1], xrange[1,1]), xrange[1,0])
                    if xconstaints is None or xconstaints( xi ):
                        break

                DeltaE = abs(objective_f(xi)-fc)
                fv = objective_f(xi)
                if np.isfinite(fv):
                    if ( fv > fc ):
                        # Initialize DeltaE_avg if a worse solution was found
                        #   on the first iteration
                        if (i==0 and j==0): DeltaE_avg = DeltaE
                        # objective function is worse
                        # generate probability of acceptance
                        p = np.exp(-DeltaE/(DeltaE_avg * t))
                        # determine whether to accept worse point
                        if (random.random()<p):
                            # accept the worse solution
                            accept = True
                        else:
                            # don't accept the worse solution
                            accept = False
                    else:
                        # objective function is lower, automatically accept
                        accept = True
                else:
                    accept = False
                if accept:
                    # update currently accepted solution
                    xc[0] = xi[0]
                    xc[1] = xi[1]
                    fc = objective_f(xc)
                    # increment number of accepted solutions
                    na = na + 1.0
                    # update DeltaE_avg
                    DeltaE_avg = (DeltaE_avg * (na-1.0) +  DeltaE) / na
            # Record the best x values at the end of every cycle
            x[i+1][0] = xc[0]
            x[i+1][1] = xc[1]
            fs[i+1] = fc
            ts[i+1] = t
            # Lower the temperature for next cycle
            t = frac * t

        self.start  = x_start
        self.n  = n
        self.x  = x
        self.fs = fs
        self.ts = ts
        self.fc = fc
        self.xc = xc


def animate_annealing( field, anneal ):
    x1m, x2m, fm = field

    # print solution
    print('Best solution: ' + str(anneal.xc))
    print('Best objective: ' + str(anneal.fc))

    # Create a contour plot
    fig = plt.figure( figsize=(15, 6) )
    gs = GridSpec( 3, 2 )

    ax0 = fig.add_subplot( gs[:, 0] )
    ax1 = fig.add_subplot( gs[0, 1] )
    ax2 = fig.add_subplot( gs[1, 1] )
    ax3 = fig.add_subplot( gs[2, 1] )

    ax0.set_title('Optimization Field')
    ax1.set_title('Convergence of the Function Value')
    ax2.set_title('Convergence of the Variable Point')
    ax3.set_title('Cooling of the Temperature ')

    # Specify contour lines
    #lines = range(2,52,2)
    # Plot contours
    CS = ax0.contour(x1m, x2m, fm)#,lines)
    # Label contours
    ax0.clabel(CS, inline=1, fontsize=10)
    # Add some text to the plot
    ax0.set_xlabel('x1')
    ax0.set_ylabel('x2')
    ax0.plot( [anneal.start[0], anneal.x[0,0]], [anneal.start[1], anneal.x[0,0]], 'y-o' )
    ax0.plot( anneal.start[0], anneal.start[1], 'ro' )
    line0, = ax0.plot(anneal.x[:,0], anneal.x[:,1], 'y-o')
    # plt.savefig('contour.png')

    line1,  = ax1.plot(anneal.fs,'r.-')
    line21, = ax2.plot(anneal.x[:,0],'b.-')
    line22, = ax2.plot(anneal.x[:,1],'g--')
    line3,  = ax3.plot(anneal.ts,'c.-')

    ax1.legend(['Objective'])
    ax2.legend(['x1','x2'])
    ax3.legend(['Temperature'])

    # Save the figure as a PNG
    # plt.savefig('iterations.png')

    # update the data
    def animate(i):
        line0.set_xdata(anneal.x[0:i,0])
        line0.set_ydata(anneal.x[0:i,1])
        line1.set_xdata(np.arange(i))
        line1.set_ydata(anneal.fs[0:i])
        line21.set_xdata(np.arange(i))
        line21.set_ydata(anneal.x[0:i, 0])
        line22.set_xdata(np.arange(i))
        line22.set_ydata(anneal.x[0:i, 1])
        line3.set_xdata(np.arange(i))
        line3.set_ydata(anneal.ts[0:i])
        return line0, line1, line21, line22, line3

    # Init only required for blitting to give a clean slate.
    def init():
        line0.set_xdata(np.ma.array(anneal.x[:,0],          mask=True))
        line0.set_ydata(np.ma.array(anneal.x[:,1],          mask=True))
        line1.set_xdata(np.ma.array(np.arange(anneal.n+1),  mask=True))
        line1.set_ydata(np.ma.array(anneal.fs,              mask=True))
        line21.set_xdata(np.ma.array(np.arange(anneal.n+1), mask=True))
        line21.set_ydata(np.ma.array(anneal.x[:,0],         mask=True))
        line22.set_xdata(np.ma.array(np.arange(anneal.n+1), mask=True))
        line22.set_ydata(np.ma.array(anneal.x[:,1],         mask=True))
        line3.set_xdata(np.ma.array(np.arange(anneal.n+1),  mask=True))
        line3.set_ydata(np.ma.array(anneal.ts,              mask=True))
        return line0, line1, line21, line22, line3

    ani = animation.FuncAnimation(fig, animate, np.arange(1, anneal.n+1),
                                    init_func=init,
                                    interval=25, blit=True)

    plt.tight_layout()
    plt.show()

def demo():

    # define objective function
    def f(x):
        x1 = x[0]
        x2 = x[1]
        obj = 0.2 + x1**2 + x2**2 - 0.1*np.cos(6.0*3.1415*x1) - 0.1*np.cos(6.0*3.1415*x2)
        return obj

    def xconstaints( x ):
        return x[0] < x[1]

    anneal = SimulatedAnnealing()
    anneal.minimize( f, xrange=[ [-1, 1], [-1, 1] ], start=[0.6, 0.6], seed=1234,
                        xconstaints=None )

    print( 'anneal.start=', anneal.start )

    # Design variables at mesh points
    i1 = np.arange(-1.0, 1.0, 0.01)
    i2 = np.arange(-1.0, 1.0, 0.01)
    x1m, x2m = np.meshgrid(i1, i2)
    fm = np.zeros(x1m.shape)
    for i in range(x1m.shape[0]):
        for j in range(x1m.shape[1]):
            fm[i][j] = 0.2 + x1m[i][j]**2 + x2m[i][j]**2 \
                 - 0.1*np.cos(6.0*3.1415*x1m[i][j]) \
                 - 0.1*np.cos(6.0*3.1415*x2m[i][j])

    animate_annealing( (x1m, x2m, fm), anneal )



# coding: utf-8
"""
    MpInverseAnim3D.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Polygon
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import mpl_toolkits.mplot3d.art3d as art3d
from matplotlib import animation
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from OurManim import manim_init
from OurMatplotlib3D import pathpatch_2d_to_3d, pathpatch_translate, plot_parallelogram3d, plot_parallelepiped
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
import molass_legacy.KekLib.DebugPlot as plt

def spike():
    manim_init()

    fig = plt.figure(figsize=(8,8))
    ax10 = fig.add_subplot(111, projection='3d')

    cube_points = [(0,0), (1,0), (1,1), (0,1)]
    cube = Polygon(cube_points, alpha=0.3, color='pink')
    ax10.add_patch(cube)
    pathpatch_2d_to_3d(cube, z=0, normal='z')
    # pathpatch_translate(cube, 0.5)

    ax10.plot([0], [0], [0], 'o', color='white')
    ax10.plot([0], [1], [0], 'o')
    ax10.plot([1], [0], [0], 'o')

    ax10.plot([1], [1], [1], 'o')
    ax10.plot([1], [-1], [2], 'o')

    M = np.array([[1, 1, 1], [1, -1, 2]]).T
    Minv = np.linalg.pinv(M)
    print('Minv=', Minv)

    p0 = np.array((0,0,0))
    p1 = np.array((1,1,1))
    p2 = np.array((1,-1,2))
    plot_parallelogram3d(ax10, p0, p1, p2, alpha=0.3, color='cyan')

    ax10.set_xlim(-2,3)
    ax10.set_ylim(-2,3)
    ax10.set_zlim(-2,3)
    fig.tight_layout()
    plt.show()

class MpInverseAnim3D(Dialog):
    def __init__(self, parent=None):
        if parent is None:
            from molass_legacy.KekLib.TkUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        self.manipulating = False
        self.mouse_axis = None
        self.in_sync = True
        self.num_frames = 85
        manim_init()
        Dialog.__init__(self, parent, "Moore-Penrose Inverse Tutorial (3D)", visible=False)

    def show(self):
        self._show()


    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = plt.figure(figsize=(18,11))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        gs = GridSpec(11,2)
        self.ax__ = self.fig.add_subplot(gs[0,:])
        self.ax00 = self.fig.add_subplot(gs[1:3,0])
        self.ax01 = self.fig.add_subplot(gs[1:3,1])
        self.ax10 = self.fig.add_subplot(gs[3:11,0], projection='3d')
        self.ax11 = self.fig.add_subplot(gs[3:11,1], projection='3d')
        # self.ax10.disable_mouse_rotation()        # to avoid a rotation bug?
        self.fig.canvas.mpl_connect( 'draw_event', self.on_draw )
        self.fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )
        self.fig.canvas.mpl_connect( 'button_release_event', self.on_button_release )

        self.draw_texts()
        self.draw()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Save", width=10, command=self.save)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw_texts(self):
        for ax in [self.ax__, self.ax00, self.ax01]:
            ax.set_axis_off()

        self.ax__.text(0.5, 0.5, "Moore-Penrose Inverse in Linear Transformation Perspective", ha='center', va='center', fontsize=40)

        self.ax00.text(0.2, 0.5, r"$ v=Mu $", ha='center', va='center', fontsize=50)
        self.ax01.text(0.18, 0.5, r"$ u=M^+v $", ha='center', va='center', fontsize=50)

        matrix00 = r"$ \begin{bmatrix} v_1 \\ v_2 \\v_3 \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \\ 1 & 2 \end{bmatrix} \begin{bmatrix} u_1 \\ u_2 \end{bmatrix} $"
        self.ax00.text(0.7, 0.5, matrix00, ha='center', va='center', fontsize=28)

        matrix01 = r"$ \begin{bmatrix} u_1 \\ u_2 \end{bmatrix} = \begin{bmatrix}  0.29 & 0.57 & 0.14 \\ 0.07 & -0.35 & 0.29 \end{bmatrix} \begin{bmatrix} v_1 \\ v_2 \\ v_3 \end{bmatrix} $"
        self.ax01.text(0.8, 0.5, matrix01, ha='center', va='center', fontsize=28)

    def draw(self):
        fig = self.fig

        ax10 = self.ax10
        ax11 = self.ax11

        u0 = np.array((0,0,0))
        u1 = np.array((1,0,0))
        u2 = np.array((0,1,0))
        p1 = np.array((1,1,1))
        p2 = np.array((1,-1,2))

        M = np.array([[1, 1, 1], [1, -1, 2]]).T
        Minv = np.linalg.pinv(M)
        print('Minv=', Minv)
        # print('Minv*p1=', np.dot(Minv, p1))
        # print('Minv*p2=', np.dot(Minv, p2))

        def set_xyzlims():
            for ax in [ax10, ax11]:
                ax.set_xlim(-2,3)
                ax.set_ylim(-2,3)
                ax.set_zlim(-2,3)

        def draw_anim(u0, u1, u2, p1, p2, v1, v2):

            square = plot_parallelogram3d(ax10, u0, u1, u2, alpha=0.3, color='pink')

            ax10.plot([0], [0], [0], 'o', color='white')
            ax10.plot([0], [1], [0], 'o')
            ax10.plot([1], [0], [0], 'o')

            ax10.plot([1], [1], [1], 'o')
            ax10.plot([1], [-1], [2], 'o')

            ax10.plot([1,1], [0,1], [0,1], ':', color='red')
            ax10.plot([0,1], [1,-1], [0,2], ':', color='blue')

            # point, = ax10.plot([u1[0]], [u1[1]], [u1[2]], 'o')
            parallel = plot_parallelogram3d(ax10, u0, v1, v2, alpha=0.3, color='cyan')
            ax10.plot(*np.array([u0,v1]).T, color='red')
            ax10.plot(*np.array([u0,v2]).T, color='blue')
            ax10.plot(*np.array([v1,v2]).T, color='orange')
            return parallel

        set_xyzlims()
        draw_anim(u0, u1, u2, p1, p2, u1, u2)

        for ax in [ax10, ax11]:
            # ax.view_init(10, 20)    # better angle
            ax.view_init(13, 25)

        u3 = np.array((0,0,1))
        q1 = np.dot(Minv, u1)
        q2 = np.dot(Minv, u2)
        q3 = np.dot(Minv, u3)
        q1_ = np.zeros(3)
        q1_[0:2] = q1
        q2_ = np.zeros(3)
        q2_[0:2] = q2
        q3_ = np.zeros(3)
        q3_[0:2] = q3

        def draw_reverse_anim(r1, r2, r3, **kwargs):
            plot_parallelepiped(ax11, (0,0,0), [r1, r2, r3], **kwargs)

        num_frames = 50
        num_moves = int(num_frames*0.8)
        w = np.linspace(0, 1, num_moves)

        def animate(i):
            ax10.cla()
            ax11.cla()
            set_xyzlims()
            if i < num_frames:
                i_ = min(num_moves-1, i)        # pause in the end of the first half
                w_ = w[i_]
            else:
                j = i - num_frames
                j_ = max(0, num_moves - 1 - j)  # pause in the end of the second half
                w_ = w[j_]
            v1 = u1*(1-w_) + p1*w_
            v2 = u2*(1-w_) + p2*w_
            draw_anim(u0, u1, u2, p1, p2, v1, v2)
            if i >= num_frames:
                draw_reverse_anim(u1, u2, u3, color='pink', edgecolor='red', alpha=0.3)
                r1 = q1_*(1-w_) + u1*w_
                r2 = q2_*(1-w_) + u2*w_
                r3 = q3_*(1-w_) + u3*w_
                draw_reverse_anim(r1, r2, r3, color='yellow', edgecolor='orange', alpha=0.3)

        self.anim = animation.FuncAnimation(fig, animate, frames=num_frames*2)

    def on_draw(self, event):
        # print('on_draw')
        if self.manipulating or self.in_sync:
            return

        this_axis = self.mouse_axis
        if this_axis == self.ax10:
            other_axis = self.ax11
        elif this_axis == self.ax11:
            other_axis = self.ax10
        else:
            other_axis = None

        if other_axis is not None:
            other_axis.view_init( this_axis.elev, this_axis.azim )

    def on_button_press(self, event):
        # print('on_button_press')
        self.manipulating = True
        self.mouse_axis = event.inaxes

    def on_button_release(self, event):
        # print('on_button_release')
        self.manipulating = False
        self.in_sync = False
        self.mouse_axis = event.inaxes

    def save(self):
        import os
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        path = os.path.join(os.getcwd(), "anim.mp4")
        self.anim.save(path, writer="ffmpeg")
        MessageBox.showinfo("Save Notification", 'The movie has been saved to "%s"' % path, parent=self)



"""
    GuiDevelopment.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk

class GuiDevelopmentMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="Development", menu=self )
        self.add_command( label="Test for input data in all subfolders in a folder", command=self.show_tester_dialog )
        self.add_command( label="Check diffences of logfiles in report folders", command=self.log_diff_dialog )
        self.add_command( label="Developer Options", command=self.show_developer_dialog )
        self.add_command( label="Restore from Result", command=self.restore_settings )
        self.add_command( label="Run Automation Script", command=self.run_automation_script )

    def show_tester_dialog(self):
        from molass_legacy.Test.TesterDialog import TesterDialog
        dialog = TesterDialog( self.parent, 'Test' )
        dialog.show()

    def log_diff_dialog(self):
        from molass_legacy.Test.TesterDialog import LogDiffDialog
        dialog = LogDiffDialog( self.parent, 'Diff logfiles' )
        dialog.show()

    def show_developer_dialog(self):
        from molass_legacy.SerialAnalyzer.DeveloperOptions import DeveloperOptionsDialog
        self.dev_dialog = DeveloperOptionsDialog( self.parent, 'Developer Options' )
        self.dev_dialog.show()

    def restore_settings(self):
        from molass_legacy.SerialAnalyzer.SettingsDialog  import RestoreSettingDialog
        dialog = RestoreSettingDialog( self.parent, 'Restore Settings' )
        if dialog.applied:
            self.parent.refresh()

    def run_automation_script(self):
        from importlib import reload
        import Test.Automation
        reload(Test.Automation)
        from molass_legacy.Test.Automation import AutomationDialog
        dialog = AutomationDialog(self.parent, self.parent)
        dialog.show()



# coding: utf-8
"""

    LambertBeerProof.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""
import numpy as np
from matplotlib.collections import PolyCollection
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from ThreeDimUtils import compute_plane
from ModeledData import ModeledData, simple_plot_3d
from .LambertBeer import BasePlane, get_base_plane, get_args
from LmfitThreadSafe import minimize, Parameters

def proof_modeled_data_demo(inset=False):
    qvector = np.linspace(0.01, 0.6, 600)
    n_elutions = 300
    pd = ModeledData(qvector, n_elutions)
    M = pd.get_data()

    index = bisect_right(qvector, 0.02)
    bp1 = get_base_plane(M, index)

    a = 0
    b = 0.1/n_elutions
    c = 0.03
    x = np.arange(len(qvector))
    y = np.arange(n_elutions)
    BP = compute_plane(a, b, c, x, y)
    M_ = M + BP
    bp2 = get_base_plane(M_, index)

    n = 180
    M1 = M_[:,0:n]
    bp3 = get_base_plane(M1, index)

    fig = plt.figure(figsize=(21,7))
    ax1 = fig.add_subplot(131, projection='3d')
    ax2 = fig.add_subplot(132, projection='3d')
    ax3 = fig.add_subplot(133, projection='3d')

    view_init = (25, -10)
    # view_init = (-10, -10)

    bp1.debug_plot(ax1, view_init=view_init)
    bp2.debug_plot(ax2, view_init=view_init)
    bp2.debug_plot(ax3, plane_color='yellow')
    bp3.debug_plot(ax3, view_init=view_init)

    y_ = np.ones(len(x))*n
    z_ = M_[:,n]
    ax3.plot(x, y_, z_, color='green', linewidth=5)

    xmin, xmax = ax3.get_xlim()
    zmin, zmax = ax3.get_zlim()
    ax3.set_xlim(xmin, xmax)
    ax3.set_zlim(zmin, zmax)

    verts = [ [ [x[0], zmin], [x[0], zmax], [x[-1], zmax], [x[-1], zmin] ] ]
    poly = PolyCollection(verts, facecolor='green', alpha=0.2)
    ax3.add_collection3d(poly, zs=[n], zdir='y')

    if inset:
        x_ = np.ones(len(y))*150
        z_ = M_[150,:]
        px1 = []
        py1 = []
        pz1 = []
        for ax in [ax1, ax2, ax3]:
            ax.plot(x_,y, z_, color='orange')

    fig.tight_layout()
    plt.show()

def proof_feasibility_demo(sd):
    from DataUtils import get_in_folder
    data = sd.intensity_array[:,:,1].T
    xray_slice = sd.xray_slice
    index = (xray_slice.start + xray_slice.stop)//2
    ecurve = sd.xray_curve
    lb = BasePlane(data, index, ecurve)

    def compute_dvector(pname, vvector):
        params = {}
        params['A'] = 0
        params['B'] = 0
        params['C'] = 0

        d_list = []
        for v in vvector:
            params[pname] = v
            d_list.append(np.sum(lb.objective(params)**2))
        dvector = np.array(d_list)
        return dvector

    a_vvector = np.linspace(-0.0002, 0.0002, 100)
    a_dvector = compute_dvector('A', a_vvector)
    b_vvector = np.linspace(-0.00005, 0.00005, 100)
    b_dvector = compute_dvector('B', b_vvector)
    c_vvector = np.linspace(-0.002, 0.002, 100)
    # c_vvector = np.linspace(-0.02, 0.02, 100)
    c_dvector = compute_dvector('C', c_vvector)

    nrows = 2
    ncols = 3
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(21,10))

    in_folder = get_in_folder()
    fig.suptitle(r"Variation of the Objective Fuction $ \| P \cdot C - M \|_{a,b,c}$ with Data " + in_folder, fontsize=30)

    for j in range(ncols):
        axes[1,j].set_yscale('log')
    for i in range(nrows):
        axes[i,0].plot(a_vvector, a_dvector)
        axes[i,1].plot(b_vvector, b_dvector)
        axes[i,2].plot(c_vvector, c_dvector)

    texts = ["a", "b", "c"]
    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i,j]
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            ax.plot([0,0], [ymin, ymax], ':', color='red')
            tx = (xmin+xmax)/2
            if i == 0:
                ty = (ymin+ymax)/2
            else:
                ty = np.sqrt(ymin*ymax)
            ax.text(tx, ty, texts[j], alpha=0.05, ha='center', va='center', fontsize=200)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()

class BaseSurfaceSpike:
    def __init__(self, data, index, ecurve):
        self.data = data
        self.index = index
        self.ecurve = ecurve

        self.compute_divided_belts()
        self.compute_soomth_surface()

    def compute_divided_belts(self):
        length = self.data.shape[0]
        belt_width = 10
        params_list = []
        for i in range(0, length, belt_width):
            start = i
            stop = min(length, i+belt_width)
            bp = BasePlane(self.data[start:stop,:], 0, self.ecurve)
            bp.solve(debug=False, debug_info=[start, stop])
            params_list.append(bp.get_params())

        if True:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            for k, i in enumerate(range(0, length, belt_width)):
                a, b, c = params_list[k]
                x = np.array([i, i])
                y = np.array([0, self.data.shape[1]-1])
                z = a*x + b*y + c
                ax.plot(x, y, z, color='red')
            fig.tight_layout()
            plt.show()

    def compute_soomth_surface(self):
        pass

def compute_basesurface(sd, pre_recog, debug=False):
    args = get_args(sd, pre_recog, debug)
    bs = BaseSurfaceSpike(*args)

def disproof_2d():
    qvector = np.linspace(0.01, 0.6, 600)
    n_elutions = 300
    pd = ModeledData(qvector, n_elutions)
    M = pd.get_data()

    index = bisect_right(qvector, 0.02)

    x = np.arange(n_elutions)
    y = M[index,:]

    A = 0.0005
    B = -0.01
    b_ = A*x + B
    y_ = y + b_

    def objective(params):
        a = params['A']
        b = params['B']
        base = A*x + B
        yc = y_ - base

        h = yc[150]
        yh = yc*h
        return yc - yh

    params = Parameters()
    params.add('A', value=0,   min=-1e-2,  max=+1e-2 )
    params.add('B', value=0,   min=-1e-2,  max=+1e-2 )

    result =  minimize( objective, params, args=() )
    A_ = result.params['A'].value
    B_ = result.params['B'].value

    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
    ax1, ax2, ax3 = axes
    for ax in axes[0:2]:
        ax.plot(x, y)
        ax.plot(x, y_)
    ax1.plot(x, b_, ':', color='red')
    ax2.plot(x, A_*x + B_, color='red')
    fig.tight_layout()
    plt.show()



"""
    ModelParams.EdmSliderInfo.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from .BaseSliderInfo import BaseSliderInfo

class EdmSliderInfo(BaseSliderInfo):
    def __init__(self, nc): 
        cmpparam_names = ["t0", "u", "a", "b", "e", "Dz", "cinj", "Rg"]

        cmpparam_indeces = []
        n_ = 7
        rg_base = nc*n_
        for k in range(nc):
            cmpparam_indeces.append(list(range(n_)) + [rg_base+k])

        BaseSliderInfo.__init__(self,
                                cmpparam_names=cmpparam_names,
                                cmpparam_indeces=cmpparam_indeces)


"""
    OurMatplotlib3D.py

    modified from
    https://stackoverflow.com/questions/18228966/how-can-matplotlib-2d-patches-be-transformed-to-3d-with-arbitrary-normals/18228967

    Copyright (c) 2018-2020, SAXS Team, KEK-PF
"""
import numpy                as np
import matplotlib.pyplot    as plt
from mpl_toolkits.mplot3d import proj3d, art3d
from matplotlib.patches import Ellipse, Polygon
from itertools import product, combinations
from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3DCollection

def rotation_matrix(d):
    """
    Calculates a rotation matrix given a vector d. The direction of d
    corresponds to the rotation axis. The length of d corresponds to 
    the sin of the angle of rotation.

    Variant of: http://mail.scipy.org/pipermail/numpy-discussion/2009-March/040806.html
    """
    sin_angle = np.linalg.norm(d)

    if sin_angle == 0:
        return np.identity(3)

    d /= sin_angle

    eye = np.eye(3)
    ddt = np.outer(d, d)
    skew = np.array([[    0,  d[2],  -d[1]],
                  [-d[2],     0,  d[0]],
                  [d[1], -d[0],    0]], dtype=np.float64)

    M = ddt + np.sqrt(1 - sin_angle**2) * (eye - ddt) + sin_angle * skew
    return M

def pathpatch_2d_to_3d(pathpatch, z = 0, normal = 'z'):
    """
    Transforms a 2D Patch to a 3D patch using the given normal vector.

    The patch is projected into they XY plane, rotated about the origin
    and finally translated by z.
    """
    if type(normal) is str: #Translate strings to normal vectors
        index = "xyz".index(normal)
        normal = np.roll((1.0,0,0), index)

    normal /= np.linalg.norm(normal) #Make sure the vector is normalised

    path = pathpatch.get_path() #Get the path and the associated transform
    trans = pathpatch.get_patch_transform()

    path = trans.transform_path(path) #Apply the transform

    pathpatch.__class__ = art3d.PathPatch3D #Change the class
    pathpatch._code3d = path.codes #Copy the codes
    pathpatch._facecolor3d = pathpatch.get_facecolor #Get the face color    

    verts = path.vertices #Get the vertices in 2D

    d = np.cross(normal, (0, 0, 1)) #Obtain the rotation vector    
    M = rotation_matrix(d) #Get the rotation matrix

    pathpatch._segment3d = np.array([np.dot(M, (x, y, 0)) + (0, 0, z) for x, y in verts])

def pathpatch_translate(pathpatch, delta):
    """
    Translates the 3D pathpatch by the amount delta.
    """
    pathpatch._segment3d += delta

def simply_to_3d(pathpatch, verts):
    path = pathpatch.get_path() #Get the path and the associated transform
    trans = pathpatch.get_patch_transform()
    path = trans.transform_path(path) #Apply the transform

    pathpatch.__class__ = art3d.PathPatch3D #Change the class
    pathpatch._code3d = path.codes #Copy the codes
    pathpatch._facecolor3d = pathpatch.get_facecolor #Get the face color    
    pathpatch._segment3d = np.array(verts)

def plot_parallelogram3d(ax, p0, p1, p2, **kwargs):
    v1 = p1 - p0
    v2 = p2 - p0
    p3 = p0 + v1 + v2
    parallel = Polygon([(0,0), (1,0), (1,1), (0,1)], **kwargs)
    ax.add_patch(parallel)
    simply_to_3d(parallel, [p0, p1, p3, p2, p0])
    return parallel

def demo_pathpatch_2d_to_3d():
    ax = plt.axes(projection = '3d') #Create axes

    p = Ellipse((0,0), .2, .5) #Add an ellipse in the yz plane
    ax.add_patch(p)
    pathpatch_2d_to_3d(p, z = 0.5, normal = 'x')
    pathpatch_translate(p, (0, 0.5, 0))

    p = Ellipse((0,0), .2, .5, facecolor = 'r') #Add an ellipse in the xz plane
    ax.add_patch(p)
    pathpatch_2d_to_3d(p, z = 0.5, normal = 'y')
    pathpatch_translate(p, (0.5, 1, 0))

    p = Ellipse((0,0), .2, .5, facecolor = 'g') #Add an ellipse in the xy plane
    ax.add_patch(p)
    pathpatch_2d_to_3d(p, z = 0, normal = 'z')
    pathpatch_translate(p, (0.5, 0.5, 0))

    for normal in product((-1, 1), repeat = 3):
        p = Ellipse((0,0), .2, .5, facecolor = 'y', alpha = .2)
        ax.add_patch(p)
        pathpatch_2d_to_3d(p, z = 0, normal = normal)
        pathpatch_translate(p, 0.5)

    plt.show()

class Inset2Din3D:
    def __init__(self, ax, inset_box):
        self.ax = ax
        self.axins = axins = ax.inset_axes(inset_box)
        axins.get_xaxis().set_visible(False)
        axins.get_yaxis().set_visible(False)

        x, y, w, h = inset_box
        b1 = ax.transAxes.transform((x, y+h))
        b2 = ax.transAxes.transform((x+w, y))

        inv = ax.transData.inverted()
        c1 = inv.transform(b1)
        c2 = inv.transform(b2)

        self.c_points = [c1, c2]
        self.annotations = []
        self.drawing = False

    def get_axis(self):
        return self.axins

    def set_event_handler(self):
        self.ax.figure.canvas.mpl_connect( 'draw_event', self.draw )

    def set_annotation_lines(self, p1, p2):
        self.a_points = [p1, p2]

    def draw(self, *args):
        ax = self.ax
        M = ax.get_proj()

        for an in self.annotations:
            an.remove()

        self.annotations = []
        for p, c in zip(self.a_points, self.c_points):
            tx, ty, _ = proj3d.proj_transform(*p, M)
            an = ax.annotate( None, xy=(tx, ty), xytext=c,
                            arrowprops=dict( headwidth=0, width=0.01, facecolor='black', shrink=0.01, alpha=0.5),
                            )
            self.annotations.append(an)

def inset_3d_demo():
    import molass_legacy.KekLib.DebugPlot as dplt

    fig = dplt.figure()
    ax = fig.add_subplot(111, projection='3d')
    theta = np.linspace(-4 * np.pi, 4 * np.pi, 100)
    z = np.linspace(-2, 2, 100)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)
    ax.plot(x, y, z, label='parametric curve')
    ax.legend(loc='upper left')

    n = 10
    px = x[n:n+1]
    py = y[n:n+1]
    pz = z[n:n+1]
    ax.plot(px, py, pz, 'o', color='red')

    inset = Inset2Din3D(ax, [0.6, 0.7, 0.4, 0.3])
    p1 = p2 = (px[0], py[0], pz[0])
    inset.set_annotation_lines(p1, p2)
    inset.draw()
    inset.set_event_handler()

    fig.tight_layout()
    dplt.show()

def parallelogram(origin, v1, v2):
    return [origin, origin+v1, origin+v1+v2, origin+v2]

def plot_parallelepiped(ax, origin, vectors, color=None, edgecolor='w', alpha=1):
    verts = []
    for i,j,k in [(0,1,2), (1,2,0), (2,0,1)]:
        vi = np.array(vectors[i])
        vj = np.array(vectors[j])
        verts.append(parallelogram(np.zeros(3), vi, vj))
        vk = np.array(vectors[k])
        verts.append(parallelogram(vk, vi, vj))
    ax.add_collection3d(Poly3DCollection(verts, color=color, linewidths=1, alpha=alpha))
    ax.add_collection3d(Line3DCollection(verts, colors=edgecolor, linewidths=0.5, linestyles=':'))

def plot_cube(ax, origin, r, **kwargs):
    pass

def plot_sphere(ax, center, r, n=100, **kwargs):
    """
        [matplotlib animation] 3.Z-axis rotation animation of the sphere
        https://sabopy.com/en/matplotlib-animation-3/
    """
    u = np.linspace(0, 2 * np.pi, n)
    v = np.linspace(0, np.pi, n)

    x, y, z = center
    x_ = r * np.outer(np.cos(u), np.sin(v)) + x
    y_ = r * np.outer(np.sin(u), np.sin(v)) + y
    z_ = r * np.outer(np.ones(np.size(u)), np.cos(v)) + z

    ax.plot_surface(x_, y_, z_, **kwargs)



"""
    Theory.RDF.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def demo_mdtraj():
    import mdtraj as md
    traj = md.load(r'D:\PyTools\Beam\1l36.pdb')
    print(traj.xyz.shape)

    pairs = traj.top.select_pairs('all', 'all')
    r, rdf_all = md.geometry.rdf.compute_rdf(traj, pairs)
    print(rdf_all[-5:])

    fig = plt.figure(figsize=(14,7))
    ax1 = fig.add_subplot(121, projection='3d')
    ax2 = fig.add_subplot(122)

    x, y, z = traj.xyz[0].T
    ax1.scatter(x, y, z)
    ax2.plot(r, rdf_all)

    fig.tight_layout()
    plt.show()

def demo2():
    domain_size = 20.0
    num_particles = 100

    particle_radius = 0.1
    rMax = domain_size / 4
    x = np.random.uniform(low=0, high=domain_size, size=num_particles)
    y = np.random.uniform(low=0, high=domain_size, size=num_particles)
    z = np.random.uniform(low=0, high=domain_size, size=num_particles)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(x, y, z)

    fig.tight_layout()
    plt.show()



"""
    RgDiscreteness.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

VERY_SMALL_VALUE = 1e-10

def compute_rg_discreteness(rgs, unit=1):
    d = max(VERY_SMALL_VALUE, np.min(np.abs(np.diff(rgs))))
    d_ = min(unit, d)
    return (unit/d_ - 1)**2

if __name__ == '__main__':
    from matplotlib import cm

    for rgs in [[70, 32, 30], [70, 31, 30], [70, 31, 30.5], [70, 31, 31]]:
        print(rgs, compute_rg_discreteness(rgs, unit=1.5))



"""
    ABC.SolverPyABC.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
import os
import logging
from importlib import reload
from molass_legacy._MOLASS.WorkUtils import get_temp_folder
from molass_legacy.Optimizer.StateSequence import save_opt_params
from molass_legacy.Optimizer.OptimizerUtils import OptimizerResult

prior_module_dict = {

}

class SolverPyABC:
    def __init__(self, optimizer):
        self.optimizer = optimizer
        self.num_pure_components = optimizer.num_pure_components
        self.cb_fh = optimizer.cb_fh
        self.nfev = 0
        self.logger = logging.getLogger(__name__)
    
    def minimize(self, objective, init_params, niter=100, seed=1234, bounds=None, callback=None):
        import Solvers.ABC.ConstrainedPrior
        reload(Solvers.ABC.ConstrainedPrior)
        from Solvers.ABC.ConstrainedPrior import ConstrainedPrior
        import Solvers.ABC.ParameterUtils
        reload(Solvers.ABC.ParameterUtils)
        from Solvers.ABC.ParameterUtils import parameter_to_vector
        import Solvers.ABC.CustomABC
        reload(Solvers.ABC.CustomABC)
        from Solvers.ABC.CustomABC import CustomABCSMC as ABCSMC
        import Solvers.ABC.HistoryUtils
        reload(Solvers.ABC.HistoryUtils)
        from Solvers.ABC.HistoryUtils import get_min_dist_parameter

        self.objective = objective

        lower = bounds[:,0]
        upper = bounds[:,1]
        constrained_prior = ConstrainedPrior(
            (self.num_pure_components, 4),  # for EGH
            lower, upper,
            offset=1,                       # for EGH
            )
       
        def objective_func_wrapper(x, x0, t, par):
            # print("objective_func_wrapper: par=", par)
            params = parameter_to_vector(par)
            fv = objective(params)
            # print("objective_func_wrapper: fv=", fv)
            return fv

        # task: remove objective_function=objective

        abc = ABCSMC(None, constrained_prior, objective_func_wrapper, population_size=200, callback=self.callback)
        db_path = os.path.join(get_temp_folder(), "abc-history.db")
        print("db_path=", db_path)
        observation = self.optimizer.xr_curve.y
        abc.new("sqlite:///" + db_path, {"data": observation})
        history = abc.run(minimum_epsilon=0.1, max_nr_populations=30)
        print("history.max_t=", history.max_t)
        parameter = get_min_dist_parameter(history)
        opt_params = parameter_to_vector(parameter)

        return OptimizerResult(x=opt_params, nit=niter, nfev=self.nfev)

    def callback(self, norm_params, f, accept):
        fv = self.objective(norm_params)
        real_params = self.optimizer.to_real_params(norm_params)
        self.logger.info("callback: fv=%.3g", fv)
        save_opt_params(self.cb_fh, real_params, fv, accept, self.nfev)
        return False



"""
    ModelParams.BaseSliderInfo.py

    Copyright (c) 20224, SAXS Team, KEK-PF
"""
import numpy as np

class BaseSliderInfo:
    def __init__(self,
                 cmpparam_names=[],
                 cmpparam_indeces=[],
                 whlparam_names=[],
                 whlparam_indeces=[]
                 ): 
        self.cmpparam_names = cmpparam_names
        self.cmpparam_indeces = np.asarray(cmpparam_indeces, dtype=int)
        self.whlparam_names = whlparam_names
        self.whlparam_indeces = np.asarray(whlparam_indeces, dtype=int)

    def get_component_indeces(self, i):
        return self.cmpparam_indeces[i,:]


# coding: utf-8
"""
    WLS_Rg.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy            as np
import statsmodels.api  as sm
from SAngler_WLS        import SAngler_WLS

class LS_Rg:
    def set_result( self, res ):
        A, B = res.params

        self.A  = A
        self.B  = B

        sigmaA, sigmaB = np.sqrt( np.diag( res.cov_params() ) )
        # print( 'sigmaA, sigmaB=', sigmaA, sigmaB )

        self.sigmaA = sigmaA
        self.sigmaB = sigmaB
        self.Rg     = np.sqrt( 3*(-B) ) if B < 0 else None
        self.I0     = np.exp( A )
        self.sigmaRg    = 1/2 / self.Rg * 3 * sigmaB if B < 0 else None
        self.sigmaI0    = self.I0 * sigmaA

class WLS_Rg( LS_Rg ):
    def __init__( self, x, y, w ):
        X = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()
        self.set_result( res )

class OLS_Rg( LS_Rg ):
    def __init__( self, x, y ):
        X = sm.add_constant(x)
        mod = sm.OLS( y, X )
        res = mod.fit()
        self.set_result( res )

class SAngler_WLS_Rg( LS_Rg ):
    def __init__( self, x, y, w ):
        mod = SAngler_WLS( y, x, w )
        res = mod.fit()
        self.set_result( res )

class ODR_WLS_Rg( LS_Rg ):
    def __init__( self, x, y, w ):
        from ODR_WLS            import ODR_WLS
        mod = ODR_WLS( y, x, w )
        res = mod.fit()
        self.set_result( res )

class WLS_HeadTrend:
    def __init__( self, x, y, w ):
        self.x  = x
        X = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        self.res = res = mod.fit()

    def compute_trend( self, guin_b, guin_size ):

        guin_angle = np.arctan2( guin_b, 1 )
        # print( 'guin_b=', guin_b, ', guin_anglee=', guin_angle )

        head_b = self.res.params[1]
        head_angle = np.arctan2( head_b, 1 ) 
        # print( 'head_b=', head_b, ', head_angle=', head_angle )

        trend_angle = ( head_angle - guin_angle ) * 180/np.pi
        # print( 'trend_angle=', trend_angle )
        trend_angle_ = max( -0.5, min( +0.5, trend_angle ) )

        head_size = self.x[-1] - self.x[0]
        proportion = min( 1, head_size/guin_size )
        # print( 'head_size, guin_size, proportion=', head_size, guin_size, proportion )

        if proportion > 0.5:
            trend = min( 1, trend_angle_ * proportion * 10 )
        else:
            # proportion==0.318 for IN_c380_00015_sub.dat
            trend = 0
        # print( 'trend=', trend )

        return trend, proportion



# coding: utf-8
"""
    UV.Spike.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import logging
import molass_legacy.KekLib.DebugPlot as plt
from .UvPreRecog import UvPreRecog

def spike(in_folder, logger, fig_file=None):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

    set_setting("in_folder", in_folder)
    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd)

    upr = UvPreRecog(sd, pre_recog, debug=True, fig_file=fig_file)
    upr.get_trim_slice()

    fc = pre_recog.flowchange
    a_curve2 = fc.a_curve2
    x = a_curve2.x
    y = a_curve2.y
    base_curve, params = upr.get_base_curve_info()
    residual_ratio = np.linalg.norm(base_curve(x, params) - y)/np.linalg.norm(y)
    return residual_ratio



"""
    Simulative.SimulationMain.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def lognormalpore_model_interactive(lrf_src, guess_info, parent=None):
    import molass_legacy.KekLib.DebugPlot as plt
    from importlib import reload
    import Simulative.LognormalPsd
    reload(Simulative.LognormalPsd)
    from Simulative.LognormalPsd import lognormalpore_model_interactive_impl

    x = lrf_src.xr_x
    y = lrf_src.xr_y
    lognormalpore_model_interactive_impl(x, y, *guess_info, use_ty_as_data=False, plot_mnp=True,
                                         window_title="Demo", parent=parent)


"""
    PreviewData.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges, shift_paired_ranges
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util

class PreviewData:
    def __init__(self, sd=None, mapper=None, judge_holder=None, paired_ranges=None,
                    xdata=None, decomp_info=None, slice_=None,
                    conc_factor=None,
                    ):
        self.logger = logging.getLogger(__name__)
        self.is_for_sec = mapper is not None
        self.mapper = mapper
        self.judge_holder = judge_holder
        self.paired_ranges = paired_ranges
        self.xdata = xdata
        self.decomp_info = decomp_info
        self.slice_ = slice_

        if self.is_for_sec:
            self.sd = sd
            self.mc_vector = sd.mc_vector
            if conc_factor is None:
                conc_factor = compute_conc_factor_util()
            conc = None
            conc_curves = None
            if paired_ranges is None:
                ranges_ = mapper.x_ranges
            else:
                ranges_ = paired_ranges
        else:
            from molass_legacy.KekLib.BasicUtils import Struct
            decomp_info = self.decomp_info
            ranges_ = decomp_info.ranges
            ft_list = ranges_[0].get_fromto_list()
            xr_j0 = ft_list[0][0]
            self.sd = Struct(xr_j0=xr_j0)   # TODO: consider to set serial_data
            self.mc_vector = np.ones(len(xdata.e_y))
            if conc_factor is None:
                conc_factor = get_setting('conc_factor')
            conc = decomp_info.conc
            conc_curves = decomp_info.curves

        ret = convert_to_paired_ranges(ranges_)
        self.cnv_ranges = ret[0]
        self.num_ranges = ret[1]

        self.logger.info("cnv_ranges=%s from ranges_=%s", self.cnv_ranges, ranges_)

        assert conc_factor is not None
        self.conc_factor = conc_factor
        self.conc = conc
        self.conc_curves = conc_curves

    def get_analysis_ranges(self):
        """
        this shift is required in MCT to make it compatible with SEC.
        """
        assert not self.is_for_sec
        return shift_paired_ranges(self.sd.xr_j0, self.cnv_ranges)

    def make_conc_vector(self):
        if self.is_for_sec:
            # conc_factor ?
            y_ = self.mapper.make_uniformly_scaled_vector(scale=1)
        else:
            y_ = self.xdata.e_curve.y
        return y_

    def __repr__(self):
        from molass_legacy._MOLASS.DummyClasses import DummySd, DummyMapper, DummyJudgeHolder
        return '%s(%s, %s, %s, %s, %s, %s, %s)' % (self.__class__.__qualname__,
                                                   DummySd(),
                                                   DummyMapper(),
                                                   DummyJudgeHolder(),
                                                   str(self.paired_ranges),
                                                   "'xdata'",
                                                   str(self.decomp_info),
                                                   str(self.slice_))

class PreviewOptions:
    def __init__(self, conc_depend=None, aq_smoothness=False, aq_positivity=False, use_elution_models=True):
        self.conc_depend = conc_depend
        self.aq_smoothness = aq_smoothness
        self.aq_positivity = aq_positivity
        self.use_elution_models = use_elution_models
    def __repr__(self):
        return 'PreviewOptions(%s, %s, %s)' % (str(self.aq_smoothness), str(self.aq_positivity), str(self.use_elution_models))
